#!/usr/bin/env python
"""Performs :term:`metagene` analyses. The workflow is separated into the
following subprograms:

Generate
    A :term:`metagene` profile is a position-wise average over all genes
    in the vicinity of an interesting landmark (e.g. a start codon). Because
    genes can have multiple transcript isoforms that may cover different
    genomic positions, which transcript positions (and therefore which
    genomic positions) to include in the average can be ambiguous when
    the isoforms are not knnow.
    
    To handle this problem, we define for each gene the :term:`maximal spanning window`
    over which every position at a given distance from the landmark of interest
    (e.g. a start or stop codon) maps to the same genomic coordinates in all
    transcript isoforms. The windows are defined by the following algorithm: 
    
     #. Transcripts are grouped by gene.
    
     #. Landmarks are detected on each transcript for each gene. For genes in
        which all transcripts do not share the same genomic coordinate for the
        landmark of interest, no window can be defined, and that gene is
        excluded from further analysis.
    
     #. For each set of transcripts that passes step (2), the maximal spanning
        window is created by aligning the set of transcripts at the landmark, and
        bidirectionally growing the maximal spanning window until either:
        
           - the next nucleotide position added no longer corresponds to 
             the same genomic position in all transcripts
            
           - the window reaches the maximum user-specified size

    **Note**: if annotations are supplied as `BED`_ files, transcripts cannot be
    grouped by gene, because `BED`_ files don't contain this information.
    In this case one ROI is generated per transcript.
    
    
    .. Rubric :: Output files
    
    OUTBASE_rois.txt
        A tab-delimited text file describing the maximal spanning window for
        each gene, with columns as follows:
        
        ================   ==================================================
        **Column**         **Contains**
        ----------------   --------------------------------------------------

        alignment_offset   Offset to align window to all other windows in the
                           file, if the window happens to be shorter on the 5'
                           end than specified in ``--flank_upstream``. Typically
                           this is `0`.

        region_id          ID of region (e.g. gene) from which window was made
        
        region             maximal spanning window, formatted as
                           `chromosome:start-end:(strand)`
        
        window_size        with of window
        
        zero_point         distance from 5' end of window to landmark
        ================   ==================================================
        
    
    OUTBASE_rois.bed
        Maximal spanning windows in `BED`_ format for visualization in
        a :term:`genome browser`. The thickly-rendered portion of a window
        indicates its landmark

    where `OUTBASE` is supplied by the user.
    
    
Count
    This program generates :term:`metagene` profiles from a dataset of
    :term:`counts` or :term:`alignments`, taking the following steps:
    
     1. The **raw counts** at each position in each :term:`maximal spanning window`
        (from the ``generate`` subprogram) fetched as a raw count vector for the
        window.

     2. A **normalized count vector** is created for each window by dividing
        its raw count vector by the total number of counts occurring within a
        user-defined normalization region within the window.
    
     3. A **metagene average** is created by taking aligning all of the
        normalized count vectors, and taking the median normalized counts
        over all vectors at each nucleotide position. Count vectors deriving
        from windows that don't meet a minimum count threshold (set via the
        ``--norm_region`` option) are excluded.
    
    
    .. Rubric :: Output files

    Raw count vectors, normalized count vectors, and metagene profiles are all
    saved as tab-delimited text files, for subsequent plotting, filtering,
    or reanalysis.
    
    OUTBASE_metagene_profile.txt
        Tab-delimited table of metagene profile, containing the following
        columns:

        ================   ==================================================
        **Column**         **Contains**
        ----------------   --------------------------------------------------
        x                  Distance in nucleotides from the landmark
        
        metagene_average   Value of metagene average at that position
        
        regions_counted    Number of maximal spanning windows included at
                           that position in the average. i.e. windows that
                           both met the threshold set by ``--min_counts`` and
                           were not masked at that position by a :term:`mask file`
        ================   ==================================================        
        
    OUTBASE_rawcounts.txt
        Saved if ``--keep`` is specified. Table of raw counts. Each row is a
        maximal spanning window for a gene, and each column a nucleotide position
        in that window. All windows are aligned at the landmark.
    
    OUTBASE_normcounts.txt
        Saved if ``--keep`` is specified. Table of normalized counts, produced
        by dividing each row in the raw count table by the of counts in that
        row within the columns specified by ``--normalize_over``.

    OUTBASE_mask.txt
        Saved if ``--keep`` is specified. Matrix of masks indicating which
        cells in ``normcounts`` were excluded from computations.

    OUTBASE_metagene_overview.[png | svg | pdf | et c...]
        Metagene average plotted above a heatmap of normalized counts,
        in which each row of pixels is a maximal spanning window for a gene,
        and rows are sorted by the column in which they reach their
        max value. To facilitate visualization, colors in the heatmap are scaled
        from 0 to the 95th percentile of non-zero numbers in the normalized counts
        
    `OUTBASE` is supplied by the user.

    
Chart
    One or more metagene profiles generated by the ``count`` subprogram,
    for example, on different datasets, are plotted against each other. 


See command-line help for each subprogram for details on parameters for each 
"""
__author__ = "joshua"

import gc
import sys
import warnings
import argparse
import inspect
import numpy
import pandas as pd
from plastid.genomics.roitools import SegmentChain, positions_to_segments
from plastid.util.io.filters import NameDateWriter
from plastid.util.io.openers import get_short_name, argsopener, NullWriter
from plastid.util.scriptlib.help_formatters import format_module_docstring
from plastid.util.services.exceptions import (
    ArgumentWarning,
    DataWarning,
    FileFormatWarning,
)
from plastid.util.scriptlib.argparsers import (
    AnnotationParser,
    AlignmentParser,
    PlottingParser,
    MaskParser,
    BaseParser,
)
from plastid.readers.bigbed import BigBedReader

warnings.simplefilter("once", DataWarning)

printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))

#===============================================================================
# helper functions to generate/handle maximum spanning windows / ROIs
#===============================================================================


def window_landmark(region, flank_upstream=50, flank_downstream=50, ref_delta=0, landmark=0):
    """Define a window surrounding a landmark in a region, if the region has such a landmark,
    (e.g. a start codon in a transcript), accounting for splicing of the region,
    if the region is discontinuous
    
    Parameters
    ----------
    transcript : |SegmentChain| or |Transcript|
        Region on which to generate a window surrounding a landmark
    
    landmark : int
        Position of the landmark within `region`

    flank_upstream : int
        Nucleotides upstream of `landmark` to include in window
    
    flank_downstream : int
        Nucleotides downstream of `landmark` to include in window
    
    ref_delta : int
        Offset from `landmark` to the reference point. If 0, the landmark
        is the reference point. Default: 0
    
    
    Returns
    -------
    |SegmentChain|
        Window of `region` surrounding landmark
    
    int
        alignment offset to the window start, if `region` itself wasn't long
        enough in the 5' direction to include the entire distance specified by
        `flank_upstream`. Use this to align windows generated around similar
        landmarks from different `regions` (e.g. windows surrounding start codons
        in various transcripts).

    (str, int, str)
        Genomic coordinate of reference point as *(chromosome name, coordinate, strand)*
    """
    if landmark + ref_delta >= flank_upstream:
        fiveprime_offset = 0
        my_start = landmark + ref_delta - flank_upstream
    else:
        fiveprime_offset = flank_upstream - landmark
        my_start = 0

    my_end = min(region.length, landmark + ref_delta + flank_downstream)
    roi = region.get_subchain(my_start, my_end)
    span = region.spanning_segment
    chrom = span.chrom
    strand = span.strand

    if landmark + ref_delta == region.length:
        if span.strand == "+":
            ref_point = (chrom, span.end, strand)
        else:
            ref_point = (chrom, span.start - 1, strand)
    else:
        ref_point = region.get_genomic_coordinate(landmark + ref_delta)

    return roi, fiveprime_offset, ref_point


def window_cds_start(transcript, flank_upstream, flank_downstream, ref_delta=0):
    """Returns a window surrounding a start codon.
    
    Parameters
    ----------
    transcript : |Transcript|
        Transcript on which to generate window
    
    flank_upstream : int
        Nucleotide length upstream of start codon to include in window,
        if `transcript` has a start codon
    
    flank_downstream : int
        Nucleotide length downstream of start codon to include in window,
        if `transcript` has a start codon
    
    ref_delta : int, optional
        Offset from  start codon to the reference point. If `0`, the landmark
        is the reference point. (Default: `0`)
    
    Returns
    -------
    |SegmentChain|
        Window surrounding start codon if `transcript` is coding. Otherwise,
        zero-length |SegmentChain| 
    
    int
        Alignment offset to the window start, if `transcript` itself wasn't long
        enough in the 5' direction to include the entire distance specified by
        `flank_upstream`. Use this to align this window to other windows generated
        around start codons in other transcripts. If transcript is not coding,
        returns :obj:`numpy.nan`

    (str, int, str)
        Genomic coordinate of reference point as *(chromosome name, coordinate, strand)*.
        If `transcript` has no start codon, returns :obj:`numpy.nan`
    """
    if transcript.cds_start is None:
        return SegmentChain(), numpy.nan, numpy.nan

    return window_landmark(
        transcript,
        flank_upstream,
        flank_downstream,
        ref_delta=ref_delta,
        landmark=transcript.cds_start
    )


def window_cds_stop(transcript, flank_upstream, flank_downstream, ref_delta=0):
    """Returns a window surrounding a stop codon.

    Parameters
    ----------
    transcript : |Transcript|
        Transcript on which to generate window
    
    flank_upstream : int
        Nucleotide length upstream of stop codon to include in window,
        if `transcript` has a stop codon
    
    flank_downstream : int
        Nucleotide length downstream of stop codon to include in window,
        if `transcript` has a stop codon
    
    ref_delta : int, optional
        Offset from  stop codon to the reference point. If `0`, the landmark
        is the reference point. (Default: `0`)
    
    Returns
    -------
    |SegmentChain|
        Window surrounding stop codon if transcript is coding. Otherwise,
        zero-length |SegmentChain| 
    
    int
        alignment offset to the window start, if `transcript` itself wasn't long
        enough in the 5' direction to include the entire distance specified by
        `flank_upstream`. Use this to align this window to other windows generated
        around stop codons in other transcripts. If transcript is not coding,
        returns :obj:`numpy.nan`

    (str, int, str)
        Genomic coordinate of reference point as *(chromosome name, coordinate, strand)*.
        If `transcript` has no stop codon, returns :obj:`numpy.nan`
    """
    if transcript.cds_start is None:
        return SegmentChain(), numpy.nan, numpy.nan

    return window_landmark(
        transcript,
        flank_upstream,
        flank_downstream,
        ref_delta=ref_delta,
        landmark=transcript.cds_end - 3
    )


def maximal_spanning_window(
        regions,
        mask_hash,
        flank_upstream,
        flank_downstream,
        window_func=window_cds_start,
        name=None,
        printer=NullWriter()
):
    """Create a maximal spanning window over `regions` surrounding a landmark,
    
    The maximal spanning window is created by:
    
     #. Applying `window_func` to each `region` in `regions` to create a sub-window
        of `region` that surrounds a landmark identified by `window_func`, 
        with up to `flank_upstream` bases 5' of the landmark, and `flank_downstream`
        bases 3` of the landmark.

     #. If the landmark in all regions corresponds to the same genomic position,
        a maximal spanning window is created by starting at the landmark,
        and growing the window in the 5' and 3' directions along all regions
        until either:
        
          - the next nucleotide position added no longer corresponds to 
            the same genomic position in all regions 
            
          - the window reaches the maximum size specified by`flank_upstream`
            (in 5' direction) or `flank_downstream` (in 3' direction)
        
    
    Parameters
    ----------
    regions : list
        List of |SegmentChains| or |Transcripts|
    
    mask_hash : |GenomeHash|
        |GenomeHash| containing regions to exclude from analysis
    
    flank_upstream : int
        Number of nucleotides upstream of landmark to include in maximal
        spanning window, if possible

    flank_downstream: int
        Number of nucleotides downstream of landmark to include in maximal
        spanning window, if possible
    
    window_func : func, optional
        Function that defines a landmark in an individual region, and builds
        a window around that landmark over that region.  As examples,
        :func:`window_cds_start` and :func:`window_cds_stop` are provided,
        though any function that meets the following criteria can be used:
        
         1. It must take the same parameters as :func:`window_cds_start`
            
         2. It must return the same types as :func:`window_cds_start`

        Such functions could choose arbitrary features as landmarks, such as
        peaks in ribosome density, nucleic acid sequence features, transcript
        start or end sites, or any property that can be deduced from a
        |Transcript|. (Default: :func:`window_cds_start`)
    
    name : str or None, optional
        Name for maximal spanning window, to which it's `ID` attribute will 
        be set. If `None`, a name will be generated. 
        
    printer : file-like, optional
        filehandle to write logging info to (Default: :func:`NullWriter`)
    
    
    Returns
    -------
    SegmentChain
        Maximal spanning window, if `regions` share the same landmark. Otherwise,
        0-length |SegmentChain|
    
    int or :obj:`numpy.nan`
        Alignment offset to the window start, if the maximal spanning window
        itself is not long enough in the 5' direction to include the entire
        distance specified by `flank_upstream`. Use this to align this window
        to other maximal spanning windows.
        
        If `regions` do not share the same landmark, :obj:`numpy.nan`
    """
    refpoints = []
    window_size = flank_upstream + flank_downstream

    # find common positions
    position_matrix = numpy.tile(numpy.nan, (len(regions), window_size))
    for n, region in enumerate(regions):
        try:
            my_roi, my_offset, genomic_refpoint = window_func(
                region, flank_upstream, flank_downstream
            )
            refpoints.append(genomic_refpoint)

            if genomic_refpoint is not numpy.nan and len(my_roi) > 0:
                pos_list = my_roi.get_position_list()  # ascending list of positions
                my_len = len(pos_list)
                assert my_offset + my_len <= window_size
                if my_roi.spanning_segment.strand == "+":
                    position_matrix[n, my_offset:my_offset + my_len] = pos_list
                else:
                    my_len = len(pos_list)
                    position_matrix[n, my_offset:my_offset + my_len] = pos_list[::-1]

        except IndexError:
            warnings.warn(
                "IndexError finding common positions at region '%s'. Ignoring region: " %
                region.get_name()
            )

    # continue only if refpoints all match
    if len(set(refpoints)) == 1 and numpy.nan not in refpoints:
        new_shared_positions = []
        if len(set(refpoints)) == 1:
            for i in range(0, position_matrix.shape[1]):
                col = position_matrix[:, i]
                if len(set(col)) == 1 and not numpy.isnan(col[0]):
                    new_shared_positions.append(int(col[0]))

        # continue only if there exist positions shared between all regions
        if len(set(new_shared_positions)) > 0:

            # define new ROI covering all positions common to all transcripts
            new_roi = SegmentChain(
                *positions_to_segments(regions[0].chrom, regions[0].strand, new_shared_positions)
            )
            if name is None:
                name = new_roi.get_name()

            new_roi.attr["ID"] = name

            if new_roi.spanning_segment.strand == "+":
                new_roi.attr["thickstart"] = genomic_refpoint[1]
                new_roi.attr["thickend"] = genomic_refpoint[1] + 1
            else:
                new_roi.attr["thickstart"] = genomic_refpoint[1]
                new_roi.attr["thickend"] = genomic_refpoint[1] + 1

            # having made sure that refpoint is same for all transcripts,
            # we use last ROI and last offset to find new offset
            # this fails if ref point is at the 3' end of the roi,
            # due to quirks of half-open coordinate systems
            # so we test it explicitly
            if flank_upstream - my_offset == my_roi.length:
                new_offset = my_offset
            else:
                zero_point_roi = new_roi.get_segmentchain_coordinate(*genomic_refpoint)
                new_offset = flank_upstream - zero_point_roi

            masks = mask_hash.get_overlapping_features(new_roi)
            mask_segs = []
            for mask in masks:
                mask_segs.extend(mask.segments)

            new_roi.add_masks(*mask_segs)

            return new_roi, new_offset

    return SegmentChain(), numpy.nan


#===============================================================================
# Subprograms
#===============================================================================


def group_regions_make_windows(
        source,
        mask_hash,
        flank_upstream,
        flank_downstream,
        window_func=window_cds_start,
        is_sorted=False,
        group_by="gene_id",
        printer=NullWriter()
):
    """Group regions of interest by a shared attribute, and generate
    maximal spanning windows for them. Results are given in a table
    suitable for use in ``count`` subprogram. Windows are generated by
    the following algorithm:

     1. Transcripts are grouped by `group_by` attribute (default: by gene).
        If all transcripts in a group share the same
        genomic coordinate for the landmark of interest (for example, if all 
        share the same start codon), then the set of transcripts is
        included in the analysis. If not, the set of transcripts and their
        associated gene are excluded from further processing.
    
     2. For each set of transcripts that pass step (1), the maximal spanning
        window is created by aligning the set of transcripts at the landmark, and
        adding nucleotide positions in transcript coordinates to the growing
        window in both 5' and 3' directions until either:
        
          - the next nucleotide position added is no longer corresponds to 
            the same genomic position in all transcripts
            
          - the window reaches the maximum user-specified size
    
    Parameters
    ----------
    source : list or generator
        Source of |SegmentChain| or |Transcript| objects, preferably with `gene_id` and
        `transcript_id` (e.g. transcripts assembled from a `GTF2`_ or `GFF3`_
        file), so that transcripts can be grouped by gene when making maximal
        spanning windows.
    
    mask_hash : |GenomeHash|
        |GenomeHash| containing regions to exclude from analysis
    
    flank_upstream : int
        Number of nucleotides upstream of landmark to include in windows
        (in transcript coordinates)

    flank_downstream: int
        Number of nucleotides downstream of landmark to include in windows
        (in transcript coordinates)
    
    window_func : func, optional
        Function that defines a landmark in an individual transcript, and builds
        a window around that landmark over that region.  As examples,
        :func:`window_cds_start` and :func:`window_cds_stop` are provided,
        though any function that meets the following criteria can be used:
        
          1. It must take the same parameters as :func:`window_cds_start`
            
          2. It must return the same types as :func:`window_cds_start`

        Such functions could choose arbitrary features as landmarks, such as
        peaks in ribosome density, nucleic acid sequence features, transcript
        start or end sites, or any property that can be deduced from a
        |Transcript|. (Default: :func:`window_cds_start`)
    
    group_by : str, optional
        Attribute by which |SegmentChains| or |Transcripts| should be grouped
        before generating maximal spanning windows (Default: `"gene_id"`)

    is_sorted : bool, optional
        Input file is sorted and/or `tabix`_-indexed. If `True`, 
        :func:`group_regions_make_windows` will take advantage of this to save memory.
        (Default: `False`)

    printer : file-like, optional
        filehandle to write logging info to (Default: :func:`NullWriter`)
    
    Returns
    -------
    :class:`pandas.DataFrame`
        A :class:`pandas.DataFrame` containing the following columns describing the
        maximal spanning windows:

            ====================   ==================================================
            *Column*               *Contains*
            --------------------   --------------------------------------------------
    
            alignment_offset       Offset to align window to all other windows in the
                                   file from the 5' end, if the window happens to be
                                   shorter on the 5' end than specified in
                                   `flank_upstream`
    
            region_id              ID of region, given by shared value of `group_by`
                                   parameter (i.e. by default is 'gene_id')
            
            region                 Maximal spanning window, formatted as
                                   `chromosome:start-end:(strand)`

            region_length          Length of maximal spanning window
            
            region_bed             Maximal spanning window, formatted as a `BED`_ line
            
            window_size            Requested length of maximal spanning window.
                                   May be larger than actual window if
                                   `alignment_offset` or `threeprime_offset` is 
                                   nonzero
            
            zero_point             Distance from 5' end of window to landmark,
                                   including `alignment_offset`
            
            threeprime_offset      Offset to align window to all other windows
                                   the file from the 3' end, if the window happens
                                   to be shorter on the 3' end than specified in
                                   `flank_downstream`
            ====================   ==================================================        
    
    list
        List of |SegmentChain| representing each window. These data are also
        represented as strings in the :class:`pandas.DataFrame`
    
    Notes
    -----         
    Not all genes will be included in the output if, for example, there isn't a
    position set common to all transcripts surrounding the landmark
    """
    import itertools
    window_size = flank_upstream + flank_downstream

    dtmp = {
        "region_id"        : [],
        "region"           : [],
        "region_length"    : [],
        "masked"           : [],
        "alignment_offset" : [],
        "window_size"      : [],
        "zero_point"       : [],
        "region_bed"       : [],
        "threeprime_offset": [],
    } # yapf: disable

    transcripts = []
    group_transcript = {}
    last_chrom = None
    do_loop = True
    source = iter(source)
    c = -1

    # to save memory, we process one chromosome at a time if input file is sorted
    # knowing that at that moment all transcript parts are assembled
    while do_loop == True:
        try:
            tx = next(source)
        except StopIteration:
            do_loop = False

        try:
            # end of chromosome or end of file
            if (is_sorted and tx.spanning_segment.chrom != last_chrom) or do_loop == False:
                last_chrom = tx.spanning_segment.chrom
                if do_loop == True:
                    source = itertools.chain([tx], source)

                for tx_chain in transcripts:
                    # if attr is missing, use transcript name, which should be unique
                    attr = tx_chain.attr
                    if group_by == "gene_id":
                        if "gene_id" in attr:
                            group_attr = attr["gene_id"]
                        else:
                            group_attr = tx_chain.get_gene()
                            warnings.warn(
                                "Region '%s' has no gene_id. Inferring gene_id to be '%s'" %
                                (tx_chain.get_name(), group_attr), DataWarning
                            )
                    else:
                        if group_by in attr:
                            group_attr = attr[group_by]
                        else:
                            warnings.warn(
                                "Region '%s' has no attribute '%s', and will not be grouped. Using region name as default group."
                                % (tx_chain.get_name(), group_by), DataWarning
                            )
                            group_attr = tx_chain.get_name()

                    try:
                        group_transcript[group_attr].append(tx_chain)
                    except KeyError:
                        group_transcript[group_attr] = [tx_chain]

                # for each gene, find maximal window in which all points
                # are represented in all transcripts. return window and offset
                for region_id, tx_list in group_transcript.items():
                    c += 1
                    if c % 1000 == 1:
                        printer.write(
                            "Processed %s genes, included %s ..." %
                            (c, len(list(dtmp.values())[0]))
                        )

                    name = region_id  # name regions after gene id
                    max_spanning_window, offset = maximal_spanning_window(
                        tx_list,
                        mask_hash,
                        flank_upstream,
                        flank_downstream,
                        window_func=window_func,
                        name=name,
                        printer=printer
                    )

                    if len(max_spanning_window) > 0:
                        mask_chain = max_spanning_window.get_masks_as_segmentchain()
                        dtmp["region_id"].append(region_id)
                        dtmp["window_size"].append(window_size)

                        # need to cast `region` to string to keep numpy from converting to array
                        dtmp["region"].append(str(max_spanning_window))
                        dtmp["masked"].append(str(mask_chain))
                        dtmp["alignment_offset"].append(offset)
                        dtmp["zero_point"].append(flank_upstream)
                        dtmp["region_bed"].append(max_spanning_window.as_bed())
                        dtmp["region_length"].append(max_spanning_window.length)
                        dtmp["threeprime_offset"
                             ].append(window_size - offset - max_spanning_window.length)

                # clean up
                del transcripts
                del group_transcript
                gc.collect()
                del gc.garbage[:]
                transcripts = []
                group_transcript = {}

            else:
                transcripts.append(tx)

        except UnboundLocalError:
            pass

    df = pd.DataFrame(dtmp)
    df.sort_values(["region_id"], inplace=True)
    printer.write("Processed %s genes total. Included %s." % (c + 1, len(df)))

    # Warn in case of annotation problems
    if (df["alignment_offset"] == flank_upstream).all():
        warnings.warn(
            "All maximal spanning windows lack flanks upstream of reference landmark. This occurs e.g. for start codons when annotation files don't contain UTR data. Please check your annotation file.",
            DataWarning
        )

    # N.b. This warning will only be invoked for zero-length landmarks
    # e.g. won't work for stop codons, which are 3nt wide
    if (df["threeprime_offset"] == flank_downstream).all():
        warnings.warn(
            "All maximal spanning windows lack flanks downstream of reference landmark. This occurs e.g. for stop codons when annotation files don't contain UTR data. Please check your annotation file.",
            DataWarning
        )

    return df


_NORM_START_DEFAULT = 20
_NORM_END_DEFAULT = 50


def _get_norm_region(roi_table, args):
    """Helper function to get normalization region from current and deprecated
    command-line arguments.  This function will be removed in plastid v0.6.1,
    when the deprecated command-line arguments will also be removed.


    Parameters
    ----------
    roi_table : :class:`pandas.DataFrame`
        DataFrame of output from |metagene| generate subprogram
        
    args : :class:`argparse.NameSpace`
        Command-line arguments to |metagene| or |psite| count subprograms
        
        
    Returns
    -------
    (int, int)
        Start and end of normalization region, with respect to start of window,
        including `alignment_offset`.
    """
    # TODO: remove --norm_region in Plastid v0.6.1
    flank_upstream = roi_table["zero_point"][0]
    if args.normalize_over is not None:
        norm_start, norm_end = args.normalize_over
        norm_start += flank_upstream
        norm_end += flank_upstream
        if args.norm_region is not None:
            warnings.warn(
                "`--normalize_over` replaces `--norm_region`, which is deprecated. Ignoring `--norm_region` and using `--normalize_over`. See `metagene count --help` for differences.",
                ArgumentWarning
            )
    elif args.norm_region is not None:
        warnings.warn(
            "`--norm_region` is deprecated and will be removed in plastid v0.5. Use `--normalize_over` instead. See `metagene count --help` for differences.",
            ArgumentWarning
        )
        norm_start, norm_end = args.norm_region
    else:
        norm_start = _NORM_START_DEFAULT
        norm_end = _NORM_END_DEFAULT

    return norm_start, norm_end


def do_count(args, alignment_parser, plot_parser, printer=NullWriter()):
    """Calculate a metagene average over maximal spanning windows specified in `roi_table`, taking the following steps:

     1. The **raw counts** at each position in each :term:`maximal spanning window`
        (from the ``generate`` subprogram) are totaled to create a raw count vector
        for the ROI.

     2. A **normalized count vector** is created fore each window by dividing
        its raw count vector by the total number of counts occurring within a
        user-defined normalization window within the window.
    
     3. A **metagene average** is created by taking aligning all of the
        normalized count vectors, and taking the median normalized counts
        over all vectors at each nucleotide position. Count vectors deriving
        from ROIs that don't meet a minimum count threshold (set via the
        ``--norm_region`` option) are excluded.
            
    Parameters
    ----------
    args : :class:`argparse.Namespace`
        Namespace containing arguments

    printer : file-like, optional
        Anything implementing a ``write()`` method, for logging purposes.

    
        
    Returns
    -------
    :py:class:`numpy.ndarray`
        raw counts at each position (column) in each window (row)
    
    :py:class:`numpy.ndarray`
        counts at each position (column) in each window (row), normalized by
        the total number of counts in that row from `norm_start` to `norm_end`
    
    :class:`pandas.DataFrame`
        Metagene profile of median normalized counts at each position across
        all windows, and the number of windows included in the calculation of each
        median
    """
    import matplotlib
    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    from plastid.plotting.plots import profile_heatmap

    plot_parser.set_style_from_args(args)

    # yapf: disable
    outbase = args.outbase
    count_fn     = "%s_rawcounts.txt.gz" % outbase
    normcount_fn = "%s_normcounts.txt.gz" % outbase
    mask_fn      = "%s_mask.txt.gz" % outbase
    profile_fn   = "%s_metagene_profile.txt" % outbase
    fig_fn       = "%s_metagene_overview.%s" % (outbase,args.figformat)
    # yapf: enable

    printer.write("Opening ROI file %s ..." % args.roi_file)
    with open(args.roi_file) as fh:
        roi_table = pd.read_table(fh, sep="\t", comment="#", index_col=None, header=0)
        fh.close()

    # wrapper to deal with multiple command-line arguments
    norm_start, norm_end = _get_norm_region(roi_table, args)

    #norm_start, norm_end = args.norm_region
    min_counts = args.min_counts

    # open count files
    ga = alignment_parser.get_genome_array_from_args(args, printer=printer)

    # following value are identical for all genes, so 0th val is fine
    window_size = roi_table["window_size"][0]
    upstream_flank = roi_table["zero_point"][0]
    cshape = (len(roi_table), window_size)

    # by default, mask everything
    counts = numpy.ma.MaskedArray(numpy.tile(numpy.nan, cshape), mask=numpy.tile(True, cshape))

    for i, row in roi_table.iterrows():
        if i % 1000 == 1:
            printer.write("Counted %s ROIs ..." % (i))

        roi = SegmentChain.from_str(row["region"])
        mask = SegmentChain.from_str(row["masked"])
        roi.add_masks(*mask)
        offset = int(round((row["alignment_offset"])))
        assert offset + roi.length <= window_size

        # take away from masked array
        mvec = roi.get_masked_counts(ga)
        counts.data[i, offset:offset + roi.length] = mvec.data
        counts.mask[i, offset:offset + roi.length] = mvec.mask
        #counts[i,offset:offset+roi.length] = roi.get_masked_counts(ga)

    printer.write("Counted %s ROIs total." % (i + 1))

    denominator = numpy.nansum(counts[:, norm_start:norm_end], axis=1)
    row_select = denominator >= min_counts

    norm_counts = (counts.T.astype(float) / denominator).T
    norm_counts = numpy.ma.MaskedArray(norm_counts, mask=counts.mask)
    norm_counts.mask[numpy.isnan(norm_counts)] = True
    norm_counts.mask[numpy.isinf(norm_counts)] = True

    if args.keep == True:
        printer.write("Saving counts to %s ..." % count_fn)
        numpy.savetxt(count_fn, counts, delimiter="\t", fmt='%.8f')
        printer.write("Saving normalized counts to %s ..." % normcount_fn)
        numpy.savetxt(normcount_fn, norm_counts, delimiter="\t")
        printer.write("Saving masks used in profile building to %s ..." % mask_fn)
        numpy.savetxt(mask_fn, norm_counts.mask, delimiter="\t")

    try:
        if args.use_mean == True:
            pfunc = numpy.ma.mean
        else:
            pfunc = numpy.ma.median

        profile = pfunc(norm_counts[row_select], axis=0)
    except IndexError:
        profile = numpy.zeros(norm_counts.shape[0])
    except ValueError:
        profile = numpy.zeros(norm_counts.shape[0])

    if profile.sum() == 0:
        printer.write(
            "Metagene profile is zero at all positions. %s ROIs made the minimum count cutoff." %
            row_select.sum()
        )
        printer.write("Consider lowering --min_counts (currently %s)." % min_counts)

    num_genes = ((~norm_counts.mask)[row_select]).sum(0)
    profile_table = pd.DataFrame(
        {
            "metagene_average": profile,
            "regions_counted": num_genes,
            "x": numpy.arange(-upstream_flank, window_size - upstream_flank),
        }
    )

    printer.write("Saving metagene profile to %s ..." % profile_fn)
    with argsopener(profile_fn, args, "w") as profile_out:
        profile_table.to_csv(
            profile_out,
            sep="\t",
            header=True,
            index=False,
            na_rep="nan",
            columns=["x", "metagene_average", "regions_counted"],
        )
        profile_out.close()

    # plot
    printer.write("Plotting to %s ..." % fig_fn)
    rs = norm_counts[row_select, :]
    p95 = numpy.nanpercentile(rs[rs > 0], 95)
    im_args = {
        "interpolation": "none",
        "vmin": 0,
        "vmax": p95,  # limit color space for better plotting
    }
    if args.cmap is not None:
        im_args["cmap"] = args.cmap

    plot_args = {}
    plot_args["color"] = plot_parser.get_colors_from_args(args, 1)[0]

    fig = plot_parser.get_figure_from_args(args)
    ax = plt.gca()
    fig, ax = profile_heatmap(
        norm_counts[row_select],
        profile=profile_table["metagene_average"],
        axes=ax,
        x=profile_table["x"],
        im_args=im_args,
        plot_args=plot_args
    )

    title = args.title if args.title is not None else "Metagene overview for %s" % outbase
    fig.suptitle(title)

    ax["main"].set_ylabel("Normalized ribosome density (au), by gene")
    landmark = args.landmark
    if args.landmark is not None:
        ax["main"].set_xlabel("Distance (nt) from %s" % landmark)

    printer.write("Saving image to %s ..." % fig_fn)
    fig.savefig(fig_fn, dpi=args.dpi, bbox_inches="tight")

    return counts, norm_counts, profile_table


def do_chart(args, plot_parser, printer=NullWriter()):
    """Plot metagene profiles against one another
    
    Parameters
    ----------
    args : :class:`argparse.Namespace`
        Namespace containing arguments

    printer : file-like, optional
        Anything implementing a ``write()`` method, for logging purposes.
    
    Returns
    -------
    :py:class:`matplotlib.Figure`
    """
    import matplotlib
    matplotlib.use("Agg")
    import matplotlib.pyplot as plt

    if len(args.labels) == len(args.infiles):
        samples = {
            K: pd.read_table(V, sep="\t", comment="#", header=0, index_col=None)
            for K, V in zip(args.labels, args.infiles)
        }
    else:
        if len(args.labels) > 0:
            warnings.warn(
                "Expected %s labels supplied for %s infiles; found only %s. Ignoring labels" %
                (len(args.infiles), len(args.infiles), len(args.labels)), ArgumentWarning
            )
        samples = {
            get_short_name(V): pd.read_table(V, sep="\t", comment="#", header=0, index_col=None)
            for V in args.infiles
        }

    plot_parser.set_style_from_args(args)
    title = args.title
    colors = plot_parser.get_colors_from_args(args, len(args.infiles))

    fig = plot_parser.get_figure_from_args(args)
    ax = plt.gca()
    min_x = numpy.inf
    max_x = -numpy.inf
    for n, (k, v) in enumerate(samples.items()):
        plt.plot(v["x"], v["metagene_average"], label=k, color=colors[n])
        min_x = min(min_x, min(v["x"]))
        max_x = max(max_x, max(v["x"]))

    plt.xlim(min_x, max_x)
    ylim = ax.get_ylim()
    plt.ylim(0, ylim[1])

    plt.xlabel("Distance from %s (nt)" % args.landmark)
    plt.ylabel("Normalized read density (au)")

    if title is not None:
        plt.title(title)

    plt.legend(bbox_to_anchor=(1.02, 1.02), loc="upper left", borderaxespad=0)

    fn = "%s.%s" % (args.outbase, args.figformat)
    printer.write("Saving to %s ..." % fn)
    fig.savefig(fn, dpi=args.dpi, bbox_inches="tight")

    return fig


#===============================================================================
# PROGRAM BODY
#===============================================================================


def main(argv=sys.argv[1:]):
    """Command-line program
    
    Parameters
    ----------
    argv : list, optional
        A list of command-line arguments, which will be processed
        as if the script were called from the command line if
        :py:func:`main` is called directly.

        Default: `sys.argv[1:]`. The command-line arguments, if the script is
        invoked from the command line
    """
    al = AlignmentParser(disabled=["normalize"])
    an = AnnotationParser()
    pp = PlottingParser()
    mp = MaskParser()
    bp = BaseParser()

    alignment_file_parser = al.get_parser()
    annotation_file_parser = an.get_parser()
    mask_file_parser = mp.get_parser()
    plotting_parser = pp.get_parser()
    base_parser = bp.get_parser()

    generator_help = "Create ROI file from genome annotation"
    generator_desc = format_module_docstring(group_regions_make_windows.__doc__)

    count_help = "Count reads falling into regions of interest, normalize, and average into a metagene profile"
    count_desc = format_module_docstring(do_count.__doc__)

    chart_help = "Plot metagene profiles"
    chart_desc = format_module_docstring(do_chart.__doc__)

    parser = argparse.ArgumentParser(
        description=format_module_docstring(__doc__),
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    subparsers = parser.add_subparsers(
        title="subcommands", description="choose one of the following", dest="program"
    )
    gparser = subparsers.add_parser(
        "generate",
        help=generator_help,
        description=generator_desc,
        parents=[base_parser, annotation_file_parser, mask_file_parser],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    cparser = subparsers.add_parser(
        "count",
        help=count_help,
        description=count_desc,
        parents=[base_parser, alignment_file_parser, plotting_parser],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    pparser = subparsers.add_parser(
        "chart",
        help=chart_help,
        description=chart_desc,
        parents=[base_parser, plotting_parser],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    # generate subprogram options
    gparser.add_argument(
        "--landmark",
        type=str,
        choices=("cds_start", "cds_stop"),
        default="cds_start",
        help="Landmark around which to build metagene profile (Default: cds_start)"
    )
    gparser.add_argument(
        "--upstream",
        type=int,
        default=50,
        help="Nucleotides to include upstream of landmark (Default: 50)"
    )
    gparser.add_argument(
        "--downstream",
        type=int,
        default=50,
        help="Nucleotides to include downstream of landmark (Default: 50)"
    )
    gparser.add_argument("--group_by",type=str,default="gene_id",
                         help="Attribute (e.g. in GTF2/GFF3 column 9) by which to group regions "+ \
                              "before generating maximal spanning windows "+ \
                              "(Default: group transcripts by gene using 'gene_id' attribute from GTF2, or 'Parent' attribute in GFF3)")
    gparser.add_argument("outbase", type=str, help="Basename for output files")

    # count subprogram options
    cparser.add_argument(
        "roi_file",
        type=str,
        help="Text file containing maximal spanning windows and offsets, " +
        "generated by the ``generate`` subprogram."
    )
    cparser.add_argument(
        "--min_counts",
        type=int,
        default=10,
        metavar="N",
        help="Minimum counts required in normalization region " +
        "to be included in metagene average (Default: 10)"
    )
    cparser.add_argument(
        "--normalize_over",
        type=int,
        nargs=2,
        metavar="N",
        default=None,
        #default=(20,50),
        help="Portion of each window against which its individual raw count profile" +
        " will be normalized. Specify two integers, in nucleotide" +
        " distance from landmark (negative for upstream, positive for downstream. Surround negative numbers with quotes.). (Default: 20 50)"
    )
    cparser.add_argument(
        "--norm_region",
        type=int,
        nargs=2,
        metavar="N",
        default=None,
        help="Deprecated. Use ``--normalize_over`` instead. " +
        "Formerly, Portion of each window against which its individual raw count profile" +
        " will be normalized. Specify two integers, in nucleotide" +
        " distance, from 5\' end of window. (Default: 70 100)"
    )
    cparser.add_argument(
        "--landmark", type=str, default=None, help="Name of landmark at zero point, optional."
    )
    cparser.add_argument(
        "--use_mean",
        default=False,
        action="store_true",
        help="If supplied, use columnwise mean to generate profile (Default: use median)"
    )
    cparser.add_argument(
        "--keep",
        default=False,
        action="store_true",
        help="Save intermediate count files. Useful for additional computations (Default: False)"
    )
    cparser.add_argument("outbase", type=str, help="Basename for output files")

    # chart subprogram arguments
    pparser.add_argument("outbase", type=str, help="Basename for output file.")
    pparser.add_argument(
        "infiles",
        type=str,
        nargs="+",
        help="One or more metagene profiles, generated by the" +
        " ``count`` subprogram, which will be plotted together."
    )
    pparser.add_argument(
        "--labels",
        type=str,
        nargs="+",
        default=[],
        help="Sample names for each metagene profile (optional)."
    )
    pparser.add_argument(
        "--landmark",
        type=str,
        default=None,
        help="Name of landmark at zero point (e.g. 'CDS start' or 'CDS stop'; optional)"
    )

    args = parser.parse_args(argv)
    bp.get_base_ops_from_args(args)

    # 'generate' subprogram
    if args.program == "generate":
        printer.write("Generating ROI file ...")
        if args.landmark == "cds_start":
            map_function = window_cds_start
        elif args.landmark == "cds_stop":
            map_function = window_cds_stop

        # check sorting
        is_sorted = (args.sorted == True) or \
                    (args.tabix == True) or \
                    (args.annotation_format == "BigBed")

        # open annotations
        printer.write("Opening annotation files: %s ..." % ", ".join(args.annotation_files))

        annotation_message = """`metagene` relies upon relationships between
        transcripts/genes to make maximal spanning windows that cover them. The
        `%s` attribute used to group these is not found in your %s file.
        Consider either (1) using a GTF2 or GFF3 file, (2) creating an extended
        BED file with this additional column, or (3) creating a BigBed file
        containing this extra column.""".replace("        ", "").replace(
            "\n", " "
        ) % (args.group_by, args.annotation_format)

        if args.annotation_format == "BED":
            if not isinstance(args.bed_extra_columns,
                              list) or args.group_by not in args.bed_extra_columns:
                warnings.warn(annotation_message, FileFormatWarning)
        elif args.annotation_format == "BigBed":
            reader = BigBedReader(args.annotation_files[0])
            if args.group_by not in reader.extension_fields:
                warnings.warn(annotation_message, FileFormatWarning)

        transcripts = an.get_transcripts_from_args(args, printer=printer)
        mask_hash = mp.get_genome_hash_from_args(args)

        # get ROIs
        printer.write("Generating regions of interest ...")
        roi_table = group_regions_make_windows(
            transcripts,
            mask_hash,
            args.upstream,
            args.downstream,
            window_func=map_function,
            printer=printer,
            is_sorted=is_sorted,
            group_by=args.group_by
        )

        roi_file = "%s_rois.txt" % args.outbase
        bed_file = "%s_rois.bed" % args.outbase
        printer.write("Saving to ROIs %s ..." % roi_file)
        with argsopener(roi_file, args, "w") as roi_fh:
            roi_table.to_csv(
                roi_fh,
                sep="\t",
                header=True,
                index=False,
                na_rep="nan",
                columns=[
                    "region_id", "window_size", "region", "masked", "alignment_offset", "zero_point"
                ]
            )
            roi_fh.close()

        printer.write("Saving BED output as %s ..." % bed_file)
        with argsopener(bed_file, args, "w") as bed_fh:
            for roi in roi_table["region_bed"]:
                bed_fh.write(roi)

            bed_fh.close()

    # 'count' subprogram
    elif args.program == "count":
        do_count(args, al, pp, printer)

    # 'plot' subprogram
    elif args.program == "chart":
        assert len(args.labels) in (0, len(args.infiles))
        do_chart(args, pp, printer)

    printer.write("Done.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.genomics.splicing"""

from plastid.genomics.roitools import SegmentChain
from plastid.genomics.splicing import get_junction_tuple  #, get_junction_name, get_junction_seq
from nose.tools import assert_equal
from nose.plugins.attrib import attr

def check_equal(a, b, msg=None):
    assert_equal(a, b, msg=msg)

#===============================================================================
# INDEX: ivc -> tuple, string, et c conversion
#===============================================================================


@attr(test="unit")
def test_get_junction_tuple():
    for txid in known_juncs:
        ivcs   = known_juncs.get(txid, [])
        tuples = known_juncs_tuples[txid] #.get(txid, [])
        yield check_equal, len(ivcs), len(tuples), "%s: segments: %d, tuple len: %d" % (txid, len(ivcs), len(tuples))

        for ivc, tup in zip(ivcs, tuples):
            yield check_get_junction_tuple, ivc, tup


def check_get_junction_tuple(ivc, expected_tup):
    """Check output of :py:func:`splice_to_tuple`
    
    Parameters
    ----------
    ivc : |SegmentChain|
         A two-exon fragment representing a splice junction

    expected_tup : tuple
        (chromosome name, half-open end of fiveprime exon, first position of threeprime exon, strand)
    """
    found_tup = get_junction_tuple(ivc)
    assert_equal(found_tup, expected_tup, "Expected %s, got %s." % (expected_tup, found_tup))


# def test_get_junction_name():
#     for txid in known_juncs:
#         ivcs   = known_juncs.get(txid,[])
#         names  = known_junc_names.get(txid,[])
#         assert_equal(len(ivcs),len(names))
#         for ivc, tup in  zip(ivcs,names):
#             yield check_get_junction_name, ivc, tup
#
# def check_get_junction_name(ivc,expected_name,flank):
#     found_name = get_junction_name(ivc)
#     assert_equal(found_name, expected_name, "Expected %s, got %s." % (expected_name, found_name))
#
# def test_get_junction_seq():
#     assert False
#
# def check_get_junction_seq():
#     assert False

known_juncs = {
    "YNL130C"         : ["YNL130C:0-53^145-180(-)"],
    "YPL249C-A"       : ["YPL249C-A:0-53^291-334(-)"],

    'YBR215W_mRNA_0'  : ['YBR215W_mRNA_0:0-108^192-2175(+)'],
    'YHL001W_mRNA_0'  : ['YHL001W_mRNA_0:0-146^544-961(+)'],
    'YIL018W_mRNA_0'  : ['YIL018W_mRNA_0:0-30^430-1280(+)'],
    'YIL133C_mRNA_0'  : ['YIL133C_mRNA_0:0-648^938-1007(-)'],
    'YIL156W_B_mRNA_0': ['YIL156W_B_mRNA_0:0-41^103-408(+)'],
    'YKL006W_mRNA_0'  : ['YKL006W_mRNA_0:0-157^555-954(+)'],
    'YMR194C_B_mRNA_0': ['YMR194C_B_mRNA_0:0-325^397-729(-)'],
    'YNL130C_mRNA_0'  : ['YNL130C_mRNA_0:0-1204^1296-1382(-)'],
    'YPL249C_A_mRNA_0': ['YPL249C_A_mRNA_0:0-415^653-697(-)'],
} # yapf: disable
known_juncs = {K: [SegmentChain.from_str(X) for X in V] for K, V in known_juncs.items()}
"""Annotated splice junctions"""

known_juncs_tuples = {
    "YNL130C"   : [("YNL130C",53,145,"-"),],
    "YPL249C-A" : [("YPL249C-A",53,291,"-"),],

    'YBR215W_mRNA_0'  : [('YBR215W_mRNA_0',108,192,'+'),],
    'YHL001W_mRNA_0'  : [('YHL001W_mRNA_0',146,544,'+'),],
    'YIL018W_mRNA_0'  : [('YIL018W_mRNA_0',30,430,'+'),],
    'YIL133C_mRNA_0'  : [('YIL133C_mRNA_0',648,938,'-'),],
    'YIL156W_B_mRNA_0': [('YIL156W_B_mRNA_0',41,103,'+'),],
    'YKL006W_mRNA_0'  : [('YKL006W_mRNA_0',157,555,'+'),],
    'YMR194C_B_mRNA_0': [('YMR194C_B_mRNA_0',325,397,'-'),],
    'YNL130C_mRNA_0'  : [('YNL130C_mRNA_0',1204,1296,'-'),],
    'YPL249C_A_mRNA_0': [('YPL249C_A_mRNA_0',415,653,'-'),],
} # yapf: disable


#!/usr/bin/env python
"""Create :term:`genome browser` tracks from :term:`read alignments`, using
:term:`mapping rules <mapping rule>` to extract the biology of interest
(e.g. ribosomal P-sites, et c) from the alignments.


Output files
------------
Tracks can be output in `wiggle`_ and `bedGraph`_ formats. Because these formats
are unstranded, two files are created:

    OUTBASE_fw.wig
        Counts at each position for the plus/forward strand of each chromosome
    
    OUTBASE_rc.wig
        Counts at each position for the minus/reverse strand of each chromosome

where `OUTBASE` is given by the user.


See also
--------
:doc:`/concepts/mapping_rules`
    Explanations of mapping rules and why they can be useful

:mod:`plastid.genomics.map_factories`
    For lists of mapping rules and their parameters
"""
__author__ = "joshua"
__date__ = "2011-03-18"
import warnings
import inspect
import sys
import argparse

from plastid.util.scriptlib.argparsers import AlignmentParser, BaseParser
from plastid.util.io.filters import NameDateWriter
from plastid.util.io.openers import get_short_name, argsopener
from plastid.plotting.colors import get_rgb255
from plastid.util.scriptlib.help_formatters import format_module_docstring

warnings.simplefilter("once")
printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))


def main(argv=sys.argv[1:]):
    """Command-line program
    
    Parameters
    ----------
    argv : list, optional
        A list of command-line arguments, which will be processed
        as if the script were called from the command line if
        :py:func:`main` is called directly.

        Default: sys.argv[1:] (actually command-line arguments)
    """
    ap = AlignmentParser()
    bp = BaseParser()
    parser = argparse.ArgumentParser(
        description=format_module_docstring(__doc__),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[bp.get_parser(), ap.get_parser()]
    )
    parser.add_argument(
        "-o",
        "--out",
        dest="outbase",
        type=str,
        required=True,
        metavar="FILENAME",
        help="Base name for output files"
    )
    parser.add_argument("--window_size",default=100000,metavar="N",type=int,
                        help="Size of nucleotides to fetch at once for export. "+\
                             "Large values are faster but require more memory "+\
                             "(Default: 100000)")

    track_opts = parser.add_argument_group(title="Browser track options")
    track_opts.add_argument(
        "--color",
        type=str,
        default=None,
        help="An RGB hex string (`'#NNNNNN'`, `N` in `[0-9,A-F]`) specifying \
                              the track color."
    )
    track_opts.add_argument(
        "-t",
        "--track_name",
        dest="track_name",
        type=str,
        help="Name to give browser track",
        default=None
    )
    track_opts.add_argument(
        "--output_format",
        choices=("bedgraph", "variable_step"),
        default="bedgraph",
        help="Format of output file (Default: bedgraph)"
    )

    args = parser.parse_args(argv)
    gnd = ap.get_genome_array_from_args(args, printer=printer)
    bp.get_base_ops_from_args(args)

    if args.track_name is None:
        name = args.outbase
    else:
        name = args.track_name

    if args.color is not None:
        fw_color = rc_color = "%s,%s,%s" % tuple(get_rgb255(args.color))
    else:
        fw_color = rc_color = "0,0,0"

    if args.output_format == "bedgraph":
        outfn = gnd.to_bedgraph
    elif args.output_format == "variable_step":
        outfn = gnd.to_variable_step

    track_fw = "%s_fw.wig" % args.outbase
    track_rc = "%s_rc.wig" % args.outbase

    with argsopener(track_fw, args, "w") as fw_out:
        printer.write("Writing forward strand track to %s ..." % track_fw)
        outfn(
            fw_out,
            "%s_fw" % name,
            "+",
            window_size=args.window_size,
            color=fw_color,
            printer=printer
        )
        fw_out.close()

    with argsopener(track_rc, args, "w") as rc_out:
        printer.write("Writing reverse strand track to %s ..." % track_rc)
        outfn(
            rc_out,
            "%s_rc" % name,
            "-",
            window_size=args.window_size,
            color=rc_color,
            printer=printer
        )
        rc_out.close()

    printer.write("Done!")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Test suite for py:mod:`plastid.plotting.colors`"""
import unittest
import numpy
from plastid.plotting.colors import get_rgb255, get_str_from_rgb255
from nose.plugins.attrib import attr


@attr(test="unit")
class TestColors(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # valid colors as hex strings and tuples
        cls.valid_colors = [
            ("#000000", (0, 0, 0)),
            ("#FFFFFF", (255, 255, 255)),
            ("#ffffff", (255, 255, 255)),
            ("#FF0000", (255, 0, 0)),
            ("#00FF00", (0, 255, 0)),
            ("#0000FF", (0, 0, 255)),
            ("#6A5ACD", (106, 90, 205)),
            ("#ee6a50", (238, 106, 80)),
            ("#bf3EFF", (191, 62, 255)),
            ("#a2CD5a", (162, 205, 90)),
        ]

        # data that should cause problems
        cls.error_colors = [
            ("#ZZZZZZ", (-5, 10, 255)),  # out of range
            ("#-23a5*", (10, 3000, 0)),  # nonsense/out of range
            ("#AA", (0, 2)),             # too short
            ("#AAAAAAA", (5, 5, 5, 5)),  # 7 digits
            ("#GG0011", (0.5, 3, 2))     # float
        ] # yapf: disable

    def test_get_rgb255_match(self):
        # test correct conversions
        for my_str, my_tup in self.valid_colors:
            self.assertEqual(tuple(get_rgb255(my_str)), my_tup)

    def test_get_rgb255_notmatch(self):
        # assert not incorrect conversions
        randidx = numpy.random.randint(0, high=len(self.valid_colors), size=len(self.valid_colors))
        for i, (my_str, _) in zip(randidx, self.valid_colors):
            if self.valid_colors[i][0].upper() != my_str.upper():
                self.assertNotEqual(tuple(get_rgb255(my_str)), self.valid_colors[i][0])

    def test_get_rgb255_knownfail(self):
        # assert invalid colors cause problems
        for my_str, _ in self.error_colors:
            self.assertRaises(ValueError, get_rgb255, my_str)

    def test_get_str_from_rgb255_match(self):
        # test correct conversions
        for my_str, my_tup in self.valid_colors:
            self.assertEqual(my_str.upper(), get_str_from_rgb255(my_tup).upper())

    def test_get_str_from_rgb255_notmatch(self):
        # assert not incorrect conversions
        randidx = numpy.random.randint(0, high=len(self.valid_colors), size=len(self.valid_colors))
        for i, (_, my_tup) in zip(randidx, self.valid_colors):
            if self.valid_colors[i][1] != my_tup:
                self.assertNotEqual(get_str_from_rgb255(my_tup), self.valid_colors[i][1])

    def test_get_str_from_rgb255_knownfail(self):
        # assert invalid colors cause problems
        for _, my_tup in self.error_colors:
            self.assertRaises(ValueError, get_str_from_rgb255, my_tup)



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.util.services.sets`"""

from plastid.util.services.sets import merge_sets
from nose.plugins.attrib import attr
from nose.tools import assert_set_equal


def check_merge_equal(expected, found):
    yield assert_set_equal, expected, found


@attr(test="unit")
def test_merge_sets(fn=merge_sets):
    """Tests :py:func:`merge_sets`"""
    a = [
        set(['h']),
        set(['a', 'b', 'm']),
        set(['c']),
        set(['c']),
        set(['b', 'j']),
        set(['i']),
        set(['k', 'o']),
        set(['n']),
        set(['g', 'o']),
        set(['i', 'k']),
        set(['e', 'i']),
        set(['a']),
        set(['d', 'h']),
        set(['l']),
        set(['g', 'j', 'k']),
        set(['f']),
        set(['b']),
        set(['g', 'n']),
        set(['d', 'm', 'p']),
        set(['i', 'n'])
    ]
    b = [
        set(['g', 'm']),
        set(['o']),
        set(['a', 'g']),
        set(['f']),
        set(['h']),
        set(['h']),
        set(['o']),
        set(['b', 'l']),
        set(['e']),
        set(['p']),
        set(['j']),
        set(['p']),
        set(['k']),
        set(['c', 'f']),
        set(['b', 'c']),
        set(['b']),
        set(['i', 'k']),
        set(['g']),
        set(['j', 'm']),
        set(['b', 'n']),
        set(['i']),
        set(['d', 'o']),
        set(['b', 'l']),
        set(['a']),
        set(['a', 'p'])
    ]
    _tform = lambda x: frozenset([frozenset(X) for X in x])
    yield check_merge_equal, _tform(fn(a)), _tform(
        [
            set(['c']),
            set(['f']),
            set(['l']),
            set(['a', 'b', 'd', 'e', 'g', 'h', 'i', 'j', 'k', 'm', 'n', 'o', 'p'])
        ]
    )
    yield check_merge_equal, _tform(fn(b)), _tform(
        [
            set(['h']),
            set(['e']),
            set(['i', 'k']),
            set(['d', 'o']),
            set(['b', 'c', 'f', 'l', 'n']),
            set(['a', 'g', 'j', 'm', 'p'])
        ]
    )



#!/usr/bin/env python
"""Function decorators, exceptions, and several commonly-used that operate
on primitive object types

Package overview
================

    ===============================================  ==============================================
    **Package module**                               **Contents**
    -----------------------------------------------  ----------------------------------------------
    :py:mod:`~plastid.util.services.decorators`         Function decorators
    :py:mod:`~plastid.util.services.exceptions`         Exceptions
    :py:mod:`~plastid.util.services.lists`              Miscellaneous list operations
    :py:mod:`~plastid.util.services.sets`               Miscellaneous set operations
    :py:mod:`~plastid.util.services.misc`               Other miscellaneous functions
    :py:mod:`~plastid.util.services.mini2to3`           Python 2.x/3.x compatability layer
    ===============================================  ==============================================
 
"""



"""Reader for `BigWig`_ files, built atop `Jim Kent's utilities`_.

.. contents::
   :local:


Summary
-------

`BigWig`_ is a binary, indexed, high-performance format that associates
quantitative values with genomic positions. Because `BigWig`_ files are
indexed, they allow both sequential and random access to data, and require
substantially less memory than unindexed `wiggle`_ or `bedGraph`_ files. Like
`wiggle`_ and `bedGraph`_ formats, `BigWig`_ is an unstranded format, so data
for plus and minus strands must be stored in separate files.


Module Contents
---------------

.. autosummary::

   BigWigReader
   BigWigIterator


Examples
--------

Fetch counts over a |Transcript|/|SegmentChain| or |GenomicSegment|::

    >>> count_data = BigWigReader("some_file.bw")

    # segment covering positions 5-2000 on chrI
    >>> segment_counts = count_data[GenomicSegment("chrI", 50, 2000, "+")]

    # a transcript- normally this would come from a BED_Reader or similar
    >>> my_transcript = Transcript(GenomicSegment("chrV", 5000, 5500, "+"),
    >>>                            GenomicSegment("chrV", 8000, 9000, "+"),
    >>>                            ID='some_transcript')
    >>> transcript_counts = count_data[my_transcript]
    >>> transcript_counts
    array([ 36, 171,  65,  95, 109, 178, 211,  71,  25, 208,  20, 189, 205,
           182, 102, 159, 154, 148,  15,  65, 237, 104, 211, 162,  22,   4,
           254,  85,  53, 160,  58,  74, 199,  85, 205, 242, 162,  23, 246,
           ...
           (rest of output omitted) ])

Efficiently fetch a :class:`numpy.ndarray` of counts covering a whole
chromosome::

    >>> chrI_counts = count_data.get_chromosome("chrI")
    >>> chrI_counts
    [ numpy array of counts covering chromosome chrI ]

Iterate over a `BigWig`_ file (this is unusual). Data are returned as tuples of
(chromosome name, start coordinate, end coordinate, and the value over those
coordinates)::

    >>> for chrom, my_start, my_end, value in count_data:
    >>>     pass # do something interesting with those values


See also
--------
:class:`~plastid.genomics.genome_array.BigWigGenomeArray`
    A GenomeArray for `BigWig`_ files.

`Kent2010 <http://dx.doi.org/10.1093/bioinformatics/btq351>`_
    Description of BigBed and BigWig formats. Especially see supplemental data.

`Source repository for Kent utilities <https://github.com/ENCODE-DCC/kentUtils.git>`_
    The header files are particularly useful.
"""
import warnings
cimport numpy
import numpy

from plastid.util.services.exceptions import DataWarning
from plastid.genomics.roitools cimport GenomicSegment, SegmentChain
from plastid.genomics.c_common cimport reverse_strand
from plastid.readers.bbifile cimport (
    _BBI_Reader,
    close_file,
    bbiSummary,
    bbiSummaryType,
    bbiSumMax,
    bbiSumMin,
    bbiSumMean,
    bbiSumCoverage,
    bbiSumStandardDeviation,
    lmInit,
    lmCleanup,
    lmAlloc,
    lm,
    bitFindClear,
)

from plastid.readers.bbifile cimport WARN_CHROM_NOT_FOUND
from plastid.genomics.c_common import _GeneratorWrapper
from plastid.util.services.mini2to3 import safe_bytes, safe_str

#===============================================================================
# INDEX: BigWig reader
#===============================================================================



cdef class BigWigReader(_BBI_Reader):
    """
    BigWigReader(filename, maxmem = 0)

    Reader providing random or sequential access to data stored in `BigWig`_
    files.

    Parameters
    ----------
    filename : str
        Name of `bigwig`_ file

    maxmem : float
        Maximum desired memory footprint for C objects, in megabytes.  May be
        temporarily exceeded if large queries are requested.  Does not include
        memory footprint of Python objects.  (Default: `0`, no limit)


    Examples
    --------

    Fetch counts over a |Transcript|/|SegmentChain| or |GenomicSegment|::

        >>> count_data = BigWigReader("some_file.bw")

        # segment covering positions 5-2000 on chrI
        >>> segment_counts = count_data[GenomicSegment("chrI", 50, 2000, "+")]

        # a transcript- normally this would come from a BED_Reader or similar
        >>> my_transcript = Transcript(GenomicSegment("chrV", 5000, 5500, "+"),
        >>>                            GenomicSegment("chrV", 8000, 9000, "+"),
        >>>                            ID='some_transcript')
        >>> transcript_counts = count_data[my_transcript]

    Efficiently fetch a :class:`numpy.ndarray` of counts covering a whole
    chromosome::

        >>> chrI_counts = count_data.get_chromosome("chrI")

    Iterate over a `BigWig`_ file (this is unusual). Data are returned as
    tuples of (chromosome name, start coordinate, end coordinate, and the value
    over those coordinates)::

        >>> for chrom, my_start, my_end, value in count_data:
        >>>     pass # do something interesting with those values

    """
    def __cinit__(self, str filename, **kwargs): #, double fill = numpy.nan):
        """Open a `BigWig`_ file.

        Parameters
        ----------
        filename : str
            Name of `bigwig`_ file

        maxmem : float
            Maximum desired memory footprint for C objects, in megabytes.  May
            be temporarily exceeded if large queries are requested.  Does not
            include memory footprint of Python objects.  (Default: 0, no limit)
            """
#         """ fill : float Value to use when there is no data covering a base
#         (e.g. zero, nan, et c. Default: `numpy.nan`) """
        self._bbifile = bigWigFileOpen(safe_bytes(filename))
        self.fill = 0.0
        self._sum = numpy.nan

    def __iter__(self):
        return _GeneratorWrapper(BigWigIterator(self, maxmem=self._maxmem), "BigWig values")

    def get(self, object roi, bint roi_order=True, double fill=numpy.nan):
        """Retrieve array of counts from a region of interest.

        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Region of interest in genome

        roi_order : bool, optional
            If `True` (default) return vector of values 5' to 3'
            relative to vector rather than genome.

        fill : double, optional
            Override fill value to put in positions with no data.
            (Default: Use value of `self.fill`)


        Returns
        -------
        :class:`numpy.ndarray`
            vector of numbers, each position corresponding to a position
            in `roi`, from 5' to 3' relative to `roi`


        See also
        --------
        plastid.genomics.roitools.SegmentChain.get_counts
            Fetch a spliced vector of data covering a |SegmentChain|
        """
        cdef:
            long start    = roi.start
            long end      = roi.end
            str  chrom    = roi.chrom
            size_t length = end - start
            double usefill = self.fill if numpy.isnan(fill) else fill

            numpy.ndarray counts = numpy.full(length, usefill, dtype=numpy.float)
            double [:] view      = counts

            lm* buf = self._get_lm()
            bbiInterval* iv
            long segstart, segend

        if isinstance(roi, SegmentChain):
            return roi.get_counts(self)

        # return empty vector if chromosome is not in BigWig file
        if chrom not in self.c_chroms():
            warnings.warn(WARN_CHROM_NOT_FOUND % (chrom, self.filename), DataWarning)
            return counts

        # populate vector
        iv = bigWigIntervalQuery(self._bbifile, safe_bytes(chrom), start, end, buf)
        while iv is not NULL:
            segstart = iv.start - start
            segend = iv.end - start
            view[segstart:segend] = iv.val
            iv = iv.next

        if roi.strand == "-" and roi_order == True:
            counts = counts[::-1]

        return counts

    def __getitem__(self, GenomicSegment roi):
        """Retrieve array of counts at each position in `roi`, in `roi`'s 5' to
        3' direction.

        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Region of interest in genome

        Returns
        -------
        :class:`numpy.ndarray`
            vector of numbers, each position corresponding to a position in
            `roi`, from 5' to 3' relative to `roi`


        See also
        --------
        plastid.genomics.roitools.SegmentChain.get_counts
            Fetch a spliced vector of data covering a |SegmentChain|
        """
        return self.get(roi)

    cdef double c_sum(self):
        """Return sum of data in `BigWig`_ file, calculating if necessary

        Returns
        -------
        double
            Sum of all values over all positions
        """
        cdef:
            double mysum = 0.0
            bigWigValsOnChrom* vals
            double * vbuf
            long length
            long i = 0

        # n.b. - we calculate the exact sum manually
        # rather than using the stored sum,
        # which is approximate
        if not numpy.isnan(self._sum):
            return self._sum
        else:
            warnings.warn("Summing over a large BigWig file can take a while.", DataWarning)
            try:
                for chrom in self.chroms:
                    vals    = self.c_get_chromosome_counts(chrom)
                    length  = vals.chromSize
                    vbuf = vals.valBuf
                    while i < length:
                        mysum += vbuf[i]
                        i += 1

                    bigWigValsOnChromFree(&vals)

                self._sum = mysum
            finally:
                bigWigValsOnChromFree(&vals)

        return mysum

    def sum(self):
        """Return sum of data in `BigWig`_ file, calculating if necessary

        Returns
        -------
        double
            Sum of all values over all positions
        """
        return self.c_sum()

    cdef bigWigValsOnChrom* c_get_chromosome_counts(self, str chrom):
        """Retrieve values across an entire chromosome.

         .. note::

            The bigWigValsOnChrom* returned by this function MUST be freed by
            you, using bigWigValsOnChromFree() to avoid memoryleaks

        Parameters
        ----------
        chrom : str
            Chromosome name

        Returns
        -------
        *bigWigValsOnChrom
            Chromosome data. chromSize and bufSize fields set to zero if the
            `chrom` is not in the `BigWig`_ file.
        """
        cdef:
            bigWigValsOnChrom* vals
            int i = 0
            bint success

        #     cdef struct bigWigValsOnChrom:
        #         bigWigValsOnChrom *next
        #         char   *chrom
        #         long   chromSize
        #         long   bufSize     # size of allocated buffer
        #         double *valBuf     # value for each base on chrom. Zero where no data
        #         Bits   *covBuf     # a bit for each base with data
        if chrom not in self.c_chroms():
            warnings.warn(WARN_CHROM_NOT_FOUND % (chrom, self.filename), DataWarning)

        vals    = bigWigValsOnChromNew()
        success = bigWigValsOnChromFetchData(vals, safe_bytes(chrom), self._bbifile)

        if success == False:
            warnings.warn(
                "Could not retrieve data for chrom '%s' from file '%s'." % (chrom, self.filename),
                DataWarning
            )
            vals.chromSize = 0
            vals.bufSize   = 0

        return vals

    def get_chromosome_counts(self, str chrom):
        """Retrieve values across an entire chromosome more efficiently than
        using ``bigwig_reader[chromosome_roi]``

        Parameters
        ----------
        chrom : str
            Chromosome name

        Returns
        -------
        :class:`numpy.ndarray`
            Numpy array of float values covering entire chromosome `chrom`.
            If `chom` is not in `BigWig`_ file, returns a numpy scalar of 0.
        """
        cdef:
            bigWigValsOnChrom* vals
            long length
            int i = 0
            numpy.ndarray counts
            numpy.double_t [:] valview
            double fill = self.fill

        #     cdef struct bigWigValsOnChrom:
        #         bigWigValsOnChrom *next
        #         char   *chrom
        #         long   chromSize
        #         long   bufSize     # size of allocated buffer
        #         double *valBuf     # value for each base on chrom. Zero where no data
        #         Bits   *covBuf     # a bit for each base with data
        try:
            vals    = self.c_get_chromosome_counts(chrom)
            length  = vals.chromSize
            if length > 0:
                valview = <numpy.double_t[:length]> vals.valBuf
                counts  = numpy.asarray(valview.copy())

                # if fill isn't 0, set no-data values in output to self.fill
                # these are in vals.covBuf
#                 if self.fill != 0:
#                     i=-1
#                     while True:
#                         i = bitFindClear(vals.covBuf, i+1, length)
#                         if i >= length: # no clear bit left
#                             break
#                         counts[i] = fill
            else:
                counts = numpy.array(0, dtype=float)
        finally:
            bigWigValsOnChromFree(&vals)

        return counts

# DISABLED - values for mean & stdev aren't even close to those expected from
# known test files. However, pulling a numpy array with get() or __getitem__
# and manually calculating mean & stdev *is* accurate
#
#     def summarize(self, GenomicSegment roi):
#         """Summarize `BigWig`_ data over the region of interest.
#
#         Note
#         ----
#         These are quickly calculated approximations. More accurate mean, stdev,
#         et c can be calculated using ``bigwig_reader[roi].mean()`` et c
#
#         Returns
#         -------
#         tuple of floats
#             `(mean, max, min, fraction of bases covered, stdev)` of data in `BigWig`_
#             file over the region defined by `roi`
#         """
#         mean  = self._summarize(roi, bbiSumMean)
#         max_  = self._summarize(roi, bbiSumMax)
#         min_  = self._summarize(roi, bbiSumMin)
#         cov   = self._summarize(roi, bbiSumCoverage)
#         stdev = self._summarize(roi, bbiSumStandardDeviation)
#
#         return (mean, max_, min_, cov, stdev)

    cdef double _summarize(self, GenomicSegment roi, bbiSummaryType type_):
        """Summarize `BigWig`_ data over ROI for a single statistic

        Parameters
        ----------
        roi : |GenomicSegment|

        type: int in [0...4]
            Type of data to calculate: mean (type_ = 0), max (1), min (2),
            fraction of bases covered (3), stdev (4)

        Returns
        -------
        float
            Summarized value
        """
        cdef:
            double retval
            str chrom = roi.chrom

        retval = bigWigSingleSummary(self._bbifile,
                                     safe_bytes(chrom),
                                     roi.start,
                                     roi.end,
                                     type_,
                                     self.fill)

        return retval

#     property fill:
#         def __set__(self, double val):
#             self.fill = val
#
#         def __get__(self):
#             return self.fill


#===============================================================================
# INDEX: BigWig iterator
#===============================================================================

####################################################################
# for some reason, the return value for this gives no repr in python
# so:
#
#     >>> z = iter(some_bigwig_file) gives
#     >>> z
#
# gives:
#
# TypeError: object of type 'getset_descriptor' has no len()
#
# while:
#
#    >>>> repr(z)
#
# gives:
#
# '<generator object at 0x7f066e0f3d98>'
#

# can't be cdef'ed or cpdef'ed due to yield
def BigWigIterator(BigWigReader reader, maxmem=0):
    """BigWigIterator(reader, maxmem = 0)

    Iterate over records in the `BigWig`_ file, sorted lexically by chromosome
    and position.

    Parameters
    ----------
    reader : |BigWigReader|
        Reader to iterate over

    maxmem : float
        Maximum desired memory footprint for C objects, in megabytes.
        May be temporarily exceeded if large queries are requested.  Does not
        include memory footprint of Python objects.  (Default: 0, no limit)


    Yields
    ------
    tuple
        `(chrom name, start, end, value)`, where start & end are zero-indexed
        and half-open

    Raises
    ------
    MemoryError
        If memory cannot be allocated
    """
    cdef:
        dict          chromsizes = reader.c_chroms()
        list          chroms     = sorted(chromsizes.keys())
        str           chrom
        long          chromlength
        double        val
        lm          * buf = NULL
        bbiInterval * iv

    for chrom in chroms:
        buf = get_lm(my_lm=buf, maxmem=maxmem)
        if not buf:
            raise MemoryError("BigWigIterator: could not allocate memory.")

        chromlength = chromsizes[chrom]
        iv = bigWigIntervalQuery(reader._bbifile,
                                 safe_bytes(chrom),
                                 0,
                                 chromlength,
                                 buf)
        while iv is not NULL:
            retval = (chrom, long(iv.start), long(iv.end), float(iv.val))
            iv = iv.next
            yield retval

    lmCleanup(&buf)



from collections import OrderedDict

from plastid.readers.bbifile cimport bbiFile, bits32, bits64, lm, lmInit, lmCleanup, freeMem
from plastid.readers.bbifile cimport _BBI_Reader
from plastid.util.services.mini2to3 import safe_bytes, safe_str
from plastid.genomics.roitools cimport GenomicSegment, SegmentChain
from plastid.genomics.c_common cimport (
    strand_to_str,
    str_to_strand,
    Strand,
    forward_strand,
    reverse_strand,
    unstranded,
    error_strand,
    _GeneratorWrapper,
)

# cdef extern from "<udc.h>":
#    cdef struct udcFile

cdef extern from "<common.h>":
    cdef struct slName:
        slName    * next
        char        name[1]

    void slFreeList(void *listPt)


cdef extern from "<bPlusTree.h>":
    cdef struct bptFile
    void bptFileDetach(bptFile **pBpt)
    void bptFileClose(bptFile **pBpt)
#     :
#         bptFile        * next
#         char           * fileName
#         struct udcFile * udc
#         bits32           blockSize
#         bits32           keySize;
#         bits32           valSize
#         bits64           itemCount
#         boolean          isSwapped
#         bits64           rootOffset

cdef extern from "<bigBed.h>":

    cdef struct bigBedInterval:
        bigBedInterval * next
        bits32           start, end
        char           * rest                #remainder of BED line
        bits32           chromId

    bbiFile* bigBedFileOpen(char* fileName)

# /* Get data for interval.  Return list allocated out of lm.  Set maxItems to maximum
#  * number of items to return, or to 0 for all items. */
    bigBedInterval * bigBedIntervalQuery(
        bbiFile *bbi,
        char *chrom,
        bits32 start,
        bits32 end,
        int maxItems,
        lm *lm
    )

# /* Convert bigBedInterval into an array of chars equivalent to what you'd get by
#  * parsing the bed file. The startBuf and endBuf are used to hold the ascii representation of
#  * start and end.  Note that the interval->rest string will have zeroes inserted as a side effect.
#  * Returns number of fields in row.  */
    int bigBedIntervalToRow(
        bigBedInterval *interval,
        char *chrom, 
        char *startBuf,
        char *endBuf,
        char **row,
        int rowSize
    )

    # number of items in file
    bits64 bigBedItemCount(bbiFile *bbi)

    # get autoSql
    char *bigBedAutoSqlText(bbiFile *bbi)

# /* Return index associated with fieldName.  Aborts if no such index.  Optionally return
#  * index in a row of this field. */
    bptFile *bigBedOpenExtraIndex(bbiFile *bbi, char *fieldName, int*retFieldIX)

# /* Return list of names of extra indexes beyond primary chorm:start-end one */
    slName *bigBedListExtraIndexes(bbiFile *bbi)

# /* Return list of intervals matching file. These intervals will be allocated out of lm. */
    bigBedInterval *bigBedNameQuery(bbiFile *bbi, bptFile *index, int fieldIx, char *name, lm *lm)

# /* Fetch all records matching any of the names. Using given index on given field.
#  * Return list is allocated out of lm. */
    bigBedInterval *bigBedMultiNameQuery(
        bbiFile *bbi,
        bptFile *index,
        int fieldIx,
        char **names,
        int nameCount,
        lm *lm
    )

cdef class BigBedReader(_BBI_Reader):
    cdef:
        bint               add_three_for_stop
        int                total_fields
        int                bed_fields
        int                num_extension_fields
        object             extension_fields  # maps name to description
        object             extension_types   # map  name to formatting func

        object    return_type
        #types.classTypes return_type

    cdef list _bigbedinterval_to_bedtext(self, bigBedInterval *iv, Strand strand=*)

    cdef _GeneratorWrapper _c_get(
        self,
        SegmentChain roi,
        bint stranded=*,
        bint check_unique=*,
        lm *my_lm=*
    )



#!/usr/bin/env python
"""This module contains paths to reference files used by various unit and funcitonal tests
"""
import os
from pkg_resources import resource_filename

#===============================================================================
# INDEX: small data set for faster/unit tests
#===============================================================================

EXTENSIONS = { "BED" : "bed",
              "GTF2" : "gtf",
              "GFF3" : "gff"
                }

MINI_PATH = resource_filename("plastid", "test/data/mini")
MINI = {
    'vec_center_0_rc.txt'  : 'count_vectors/center_0_rc.txt',
    'vec_center_12_rc.txt' : 'count_vectors/center_12_rc.txt',
    'vec_center_0_fw.txt'  : 'count_vectors/center_0_fw.txt',
    'vec_center_12_fw.txt' : 'count_vectors/center_12_fw.txt',

    'vec_fiveprime_0_fw.txt'  : 'count_vectors/fiveprime_0_fw.txt',
    'vec_fiveprime_0_rc.txt'  : 'count_vectors/fiveprime_0_rc.txt',
    'vec_fiveprime_15_fw.txt' : 'count_vectors/fiveprime_15_fw.txt',
    'vec_fiveprime_15_rc.txt' : 'count_vectors/fiveprime_15_rc.txt',

    'vec_threeprime_0_rc.txt'  : 'count_vectors/threeprime_0_rc.txt',
    'vec_threeprime_0_fw.txt'  : 'count_vectors/threeprime_0_fw.txt',
    'vec_threeprime_15_fw.txt' : 'count_vectors/threeprime_15_fw.txt',
    'vec_threeprime_15_rc.txt' : 'count_vectors/threeprime_15_rc.txt',

    'fasta_file'          : 'fasta/chrA.fa',
    'bamfile'             : 'align/chrA_tophat.bam',
    'bowtie_file'         : 'align/chrA_unspliced.bowtie',
    'bed_file'            : 'bed/chrA.bed',
    'bigbed_file'         : 'bed/chrA.bb',
    'gtf2_file'           : 'bed/chrA.gtf',
    'gff3_file'           : 'bed/chrA.gff',
    'psl_file'            : 'bed/chrA.psl',

    'wig_bedgraph'         : 'wig/bedgraph',
    'wig_bedgraph_fw'      : 'wig/bedgraph_fw.wig',
    'wig_bedgraph_rc'      : 'wig/bedgraph_rc.wig',

    'bigwig'               : 'wig/bedgraph',
    'bigwig_fw'            : 'wig/bigwig_fw.bw',
    'bigwig_rc'            : 'wig/bigwig_rc.bw',

    
    'wig_variable_step'    : 'wig/variable_step',
    'wig_variable_step_fw' : 'wig/variable_step_fw.wig',
    'wig_variable_step_rc' : 'wig/variable_step_rc.wig',
} # yapf: disable

MINI = {K: os.path.join(MINI_PATH, V) for K, V in MINI.items()}
"""Dictionary of paths to files describing a small set of 28-mer alignments"""

#===============================================================================
# INDEX: big data set for functional tests
#===============================================================================

RPATH = resource_filename("plastid", "test/data")

REF_FILES = {
    # full yeast annotation
    "yeast_gtf":
    os.path.join(RPATH, "annotations", "sgd_plus_utrs.gtf"),
    "yeast_gff":
    os.path.join(RPATH, "annotations", "sgd_plus_utrs.gff"),

    # 100 yeast transcripts, used in testing various readers
    "100transcripts_gtf"  :  os.path.join(RPATH,"annotations","100transcripts.gtf"),
    "100transcripts_gff"  :  os.path.join(RPATH,"annotations","100transcripts.gff"),
    "100transcripts_bed"  :  os.path.join(RPATH,"annotations","100transcripts.bed"),
    "100cds_bed"          :  os.path.join(RPATH,"annotations","100cds.bed"),
    "100cds_antisense_bed":  os.path.join(RPATH,"annotations","100cds_antisense.bed"),

    "100transcripts_tabixbed" :  os.path.join(RPATH,"annotations","100transcripts_sorted.bed.gz"),
    "100transcripts_bigbed"   :  os.path.join(RPATH,"annotations","100transcripts.bb"),
    "100cds_bigbed"           :  os.path.join(RPATH,"annotations","100cds.bb"),
    "100cds_antisense_bigbed" :  os.path.join(RPATH,"annotations","100cds_antisense.bb"),

    "100transcripts_gtf_tabix" : os.path.join(RPATH,"annotations","100transcripts_sorted.gtf.gz"),
    "100transcripts_bed_tabix" : os.path.join(RPATH,"annotations","100transcripts_sorted.bed.gz"),

    "100cds_bed_tabix"    : os.path.join(RPATH,"annotations","100cds_sorted.bed.gz"),
    "100as_cds_bed_tabix" : os.path.join(RPATH,"annotations","100cds_antisense_sorted.bed.gz"),

    # reference output files from command-line scripts
    "yeast_parent_child"  : os.path.join(RPATH,"annotations","sgd_plus_utrs_gff3_parent_child_table.txt"),
    "yeast_parent_child_exclude_cols"  : os.path.join(RPATH,"annotations","sgd_plus_utrs_gff3_parent_child_table_exclude.txt"),
    "yeast_fasta"         : os.path.join(RPATH,"annotations","2013-01-23.saccharomyces_cerevisiae.fa"),
    "yeast_twobit"        : os.path.join(RPATH,"annotations","2013-01-23.saccharomyces_cerevisiae.2bit"),

    "yeast_juncs"         : os.path.join(RPATH,"command_line","sgd_plus_utrs.juncs"),
    "yeast_crossmap_o12_26_0"    : os.path.join(RPATH,"command_line","2013-01-23.sc_o12_26_0_crossmap.bed"),
    "yeast_crossmap_o12_26_0_bb" : os.path.join(RPATH,"command_line","2013-01-23.sc_o12_26_0_crossmap_sorted.bb"),

    "yeast_crossmap_o0_26_0"    : os.path.join(RPATH,"command_line","2013-01-23.sc_o0_26_0_crossmap.bed"),
    "yeast_crossmap_o12_26_2"   : os.path.join(RPATH,"command_line","2013-01-23.sc_o12_26_2_crossmap.bed"),

    "yeast_rp_bam"          : os.path.join(RPATH,"command_line","gen_reads.bam"),
    "yeast_psite"           : os.path.join(RPATH,"command_line","gen_p_offsets.txt"),
    "yeast_psite_constrain" : os.path.join(RPATH,"command_line","gen_p_offsets_constrain.txt"),

    "yeast_mini_bed"        : os.path.join(RPATH,"command_line","transcripts_chosen.bed"),
    "yeast_phasing"         : os.path.join(RPATH,"command_line","gen_phasing.txt"),
    "yeast_phasing_roifile" : os.path.join(RPATH,"command_line","gen_phasing_roifile.txt"),

    "yeast_metagene_cds_start" : os.path.join(RPATH,"command_line","gen_cds_start_rois.txt"),
    "yeast_metagene_cds_stop"  : os.path.join(RPATH,"command_line","gen_cds_stop_rois.txt"),              
    "yeast_metagene_cds_start_bed" : os.path.join(RPATH,"command_line","gen_cds_start_rois.bed"),
    "yeast_metagene_cds_stop_bed"  : os.path.join(RPATH,"command_line","gen_cds_stop_rois.bed"),

    "yeast_metagene_cds_start_profile" : os.path.join(RPATH,"command_line","gen_cds_start_metagene_profile.txt"),
    "yeast_metagene_cds_stop_profile"  : os.path.join(RPATH,"command_line","gen_cds_stop_metagene_profile.txt"),

    "yeast_metagene_cds_start_normcounts" : os.path.join(RPATH,"command_line","gen_cds_start_normcounts.txt"),
    "yeast_metagene_cds_stop_normcounts"  : os.path.join(RPATH,"command_line","gen_cds_stop_normcounts.txt"),

    "yeast_metagene_cds_start_rawcounts" : os.path.join(RPATH,"command_line","gen_cds_start_rawcounts.txt"),
    "yeast_metagene_cds_stop_rawcounts"  : os.path.join(RPATH,"command_line","gen_cds_stop_rawcounts.txt"),

    "yeast_findjuncs_bed"   :  os.path.join(RPATH,"annotations","sgd_plus_utrs_findjuncs.bed"),
    "yeast_findjuncs_juncs" :  os.path.join(RPATH,"annotations","sgd_plus_utrs_findjuncs.juncs"),
    "yeast_findjuncs_top"   :  os.path.join(RPATH,"annotations","sgd_plus_utrs_findjuncs_top250.bed"),
    "yeast_findjuncs_bot" :  os.path.join(RPATH,"annotations","sgd_plus_utrs_findjuncs_bot230.bed"),

    "yeast_counts_in_region_mask"    : os.path.join(RPATH,"command_line","gen_counts_in_region_mask.txt"),
    "yeast_counts_in_region_no_mask" : os.path.join(RPATH,"command_line","gen_counts_in_region_no_mask.txt"),
    "yeast_counts_in_region_sum_1" : os.path.join(RPATH,"command_line","gen_counts_in_region_sum_1.txt"),


    'slidejuncs_seqs'                        : os.path.join(RPATH,'annotations','juncs','junction_seqs.fa'),
    'slidejuncs_input'                       : os.path.join(RPATH,'annotations','juncs','input.bed'),
    'slidejuncs_ref'                         : os.path.join(RPATH,'annotations','juncs','all_known_ref.bed'),
    'slidejuncs_crossmap'                    : os.path.join(RPATH,'annotations','juncs','junc_crossmap.bed'),
    'slidejuncs_known_juncs_non_crossmap'    : os.path.join(RPATH,'annotations','juncs','known_juncs_non_crossmapping.bed'),
    'slidejuncs_known_juncs_crossmap'        : os.path.join(RPATH,'annotations','juncs','known_juncs_crossmapping.bed'),
    'slidejuncs_to_slide_known_non_crossmap' : os.path.join(RPATH,'annotations','juncs','to_slide_non_crossmapping.bed'),
    'slidejuncs_to_slide_known_crossmap'     : os.path.join(RPATH,'annotations','juncs','to_slide_crossmapping.bed'),
    'slidejuncs_noncan_no_ref'               : os.path.join(RPATH,'annotations','juncs','noncan_no_ref.bed'),
    'slidejuncs_expected_untouched'          : os.path.join(RPATH,'annotations','juncs','expected_untouched.bed'),
    
} # yapf: disable

COUNT_OPTIONS = " --count_files %s --fiveprime_variable --offset %s" % (REF_FILES["yeast_rp_bam"],REF_FILES["yeast_psite"])
ANNOTATION_OPTIONS = " --annotation_format GTF2 --add_three --annotation_files %s" % REF_FILES["yeast_gtf"]
MASK_OPTIONS = " --mask_annotation_format BED --mask_annotation_files %s" % REF_FILES["yeast_crossmap_o12_26_0"]



#!/usr/bin/env python
"""A single reader for fixedStep `wiggle`_, variableStep `wiggle`_, and `bedGraph`_ files.
|WiggleReader| is seldom called directly. Typically it is internally called by
|GenomeArray| and |SparseGenomeArray|, when their
:meth:`~plastid.genomics.genome_array.GenomeArray.add_from_wiggle` methods
are called.

See also
--------
`UCSC file format FAQ <http://genome.ucsc.edu/FAQ/FAQformat.html>`_
    UCSC Wiggle and bedGraph file specification

|GenomeArray| and |SparseGenomeArray|
    Array-like objects that store and index quantitative data over genomes
"""


class WiggleReader(object):
    """Read `wiggle`_ and `bedGraph`_ files line-by-line, returning tuples
    of `(chromosome, start position, stop position, value)`. Tuple 
    coordinates are zero-indexed and half-open, regardless of whether
    the file is a `wiggle`_ or `bedGraph`_.

    See the `UCSC file format FAQ <http://genome.ucsc.edu/FAQ/FAQformat.html>`_ for details.
    
    Parameters
    ----------
    fh : file-like
        Open filehandle pointing to `wiggle`_ or `bedGraph`_ data    
    """

    def __init__(self, fh):
        """
        Parameters
        ----------
        fh : file-like
            Open filehandle pointing to `wiggle`_ or `bedGraph`_ data
        """
        self.fh = fh
        self.data_format = "bedGraph"
        self._reset()

    def _reset(self):
        """Reset positional counters"""
        self.chrom = None
        self.step = 1
        self.span = 1  #default span is 1 (e.g. things span 1 base)
        self.counter = 1  #chromosomal positions are 1-indexed

    def __iter__(self):
        return self

    def _get_lineinfo(self, line):
        """Determine line type and return a dictionary containing key-value
        pairs of any parameters defined in the line.

        For data type declarations, the file format is returned under the key
        `'line_type'`. For track declaration lines, whatever key-value pairs
        are found are returned. 
        
        Parameters
        ----------
        line : str
            A line of `Wiggle`_ or `bedGraph`_ file
        
        Returns
        -------
        dict
            `'line-type'` key describes the type of line

            If a header line, also contains values for the current
            `'stepsize'` and `'span'`
        """
        dReturn = {}
        items = line.split()
        line_type = None
        if items[0] == "track":
            line_type = "fileHeader"
        elif items[0] == "variableStep":
            line_type = "dataHeader"
        elif items[0] == "fixedStep":
            line_type = "dataHeader"
        elif len(items) == 4:
            line_type = "bedGraph"
        else:
            line_type = "data"
        dReturn["line_type"] = line_type
        if "Header" in line_type:
            for item in items[1:]:
                if "description" in item:
                    break
                elif "=" in item:
                    key, val = item.split("=")
                    dReturn[key] = val
                else:
                    dReturn[key] = 'true'
        return dReturn

    def _next_line(self):
        return next(self.fh)

    def __next__(self):
        return self.next()

    def next(self):
        """Yield a tuple of `(chromosome, start, stop, value)`  for each data line.
        Header lines are processed internally and not exposed to the user.
        
        All coordinates are returned as 0-based, half-open intervals,
        following Python conventions.
                      
        Returns
        -------
        str
            chromosome name

        int
            start position, 0-indexed

        int
            end position, 0-indexed, half-open

        float
            value on chromosome between start and end
        """
        while True:
            line = self._next_line()
            if line.isspace():
                continue
            if line[0] == "#":
                continue

            line_info = self._get_lineinfo(line)
            line_type = line_info["line_type"]
            line_items = line.split()

            if line_type == "fileHeader":
                self.file_info = line_info
                continue
            elif line_type == "bedGraph":
                # bedGraph is 0-indexed, half open
                self._reset()
                self.data_format = "bedGraph"
                chrom = line_items[0]
                start = int(line_items[1])
                stop = int(line_items[2])
                val = float(line_items[3])
                return (chrom, start, stop, val)
            elif line_type == "dataHeader":
                self._reset()
                self.data_format = line_items[0]
                if "chrom" in line_info:
                    self.chrom = line_info["chrom"]
                if "span" in line_info:
                    self.span = int(line_info["span"])
                if "step" in line_info:
                    self.step = int(line_info["step"])
                if "start" in line_info:
                    self.counter = int(line_info["start"])
                continue
            elif line_type == "data":
                # variableStep and fixedStep are 1-indexed
                # variableStep is also fully-closed
                if self.data_format == "variableStep":
                    start = int(line_items[0]) - 1
                    stop = start + self.span  #leave alone. this will make half-open interval
                    val = float(line_items[1])
                    return (self.chrom, start, stop, val)
                if self.data_format == "fixedStep":
                    start = self.counter - 1
                    stop = start + self.span
                    val = float(line.strip())
                    self.counter += self.step
                    return (self.chrom, start, stop, val)



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.cs`"""
import tempfile
import os

from nose.plugins.attrib import attr
from pkg_resources import resource_filename, cleanup_resources
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import (
    COUNT_OPTIONS,
    ANNOTATION_OPTIONS,
    MASK_OPTIONS,
)
from plastid.bin.cs import main

from plastid.util.services.decorators import catch_stderr

#===============================================================================
# INDEX: global constants used by tests
#===============================================================================

test_info = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.cs",
    "ref_file_path": resource_filename("plastid", "test/data/command_line"),
    "temp_file_path": tempfile.mkdtemp(prefix="cs"),
}
"""Constants used by multiple tests"""

_outbase = os.path.join(test_info["temp_file_path"], "test_cs")
_file_stubs = [
    'merged.txt',
    'gene.positions',
    'transcript.positions',
    'gene_cds.bed',
    'gene_exon.bed',
    'gene_masked.bed',
    'gene_utr3.bed',
    'gene_utr5.bed',
    'transcript_cds.bed',
    'transcript_exon.bed',
    'transcript_masked.bed',
    'transcript_utr3.bed',
    'transcript_utr5.bed',
]

#===============================================================================
# INDEX: tests to execute
#===============================================================================

cs_tests = [
    (
        "generate %s %s %s" % (_outbase, ANNOTATION_OPTIONS, MASK_OPTIONS),
        [os.path.join(test_info["ref_file_path"], "gen_cs_" + X) for X in _file_stubs],
        [os.path.join(test_info["temp_file_path"], "test_cs_" + X) for X in _file_stubs], [
            "--no_header",
            "--sort_keys region",
            "--sort_keys region",
        ] + ["--no_header --sort_keys 3"] * 10
    ),
    (
        "count %s_gene.positions %s_count %s " % (_outbase, _outbase, COUNT_OPTIONS),
        [os.path.join(test_info["ref_file_path"], 'gen_cs_count_gene.txt')],
        [os.path.join(test_info["temp_file_path"], "test_cs_count.txt")], ["--sort_keys region"]
    ),
    (
        "count %s_gene.positions %s_count %s --sum 1e9" % (_outbase, _outbase, COUNT_OPTIONS),
        [os.path.join(test_info["ref_file_path"], 'gen_cs_count_gene_sum_1.txt')],
        [os.path.join(test_info["temp_file_path"], "test_cs_count.txt")], ["--sort_keys region"]
    ),
    #("",[],[],[]),
]
"""Functional tests of :py:mod:`plastid.bin.cs`.

Tests are specified as tuples of:

    1. Command-line style arguments to pass to :py:func:`main`

    2. A list of reference files that output should be compared against

    3. A list of output files created by running :py:func:`main`
       with the arguments provided in (1)

    4. A list of strings specifying how equality should be evaluated
"""

#===============================================================================
# INDEX: test/helper functions
#===============================================================================


@attr(test="functional")
@attr(speed="slow")
def do_test():
    """Perform functional test for :py:mod:`plastid.bin.cs`"""
    for x in execute_helper(test_info, cs_tests):
        yield x



#!/usr/bin/env python
"""Count the number of :term:`read alignments<alignment>` and calculate
read densities (in :term:`RPKM`) over genes, correcting gene boundaries
for overlap with other genes or regions specified in a :term:`mask file`.

:term:`Counts <counts>` and densities are calculated separately per gene for
exons, 5' UTRs, coding regions, and 3' UTRs. In addition, positions overlapped
by multiple genes are excluded, as are positions annotated in
:term:`mask annotation files<crossmap>`, if one is provided.

The script's operation is divided into three subprograms:

Generate
    The :func:`generate <do_generate>` mode pre-process a genome annotation
    as follows:

     #. All genes whose transcripts share exact exons are collapsed to
        "merged" genes.

     #. Positions covered by more than one merged gene on the same strand
        are excluded from analysis, and subtracted from each merged genes.

     #. Remaining positions in each merged gene are then divided
        into the following groups:

            *exon*
                all positions in all transcripts mapping to the merged gene

            *CDS*
                positions which appear in coding regions in *all* transcript
                isoforms mapping to the merged gene. i.e. These positions
                are never part of a fiveprime or threeprime UTR in *any*
                transcript mapping to the merged gene

            *UTR5*
                positions which are annotated only as *5' UTR* in all
                transcript isoforms mapping to the merged gene

            *UTR3*
                positions which are annotated only as *3 UTR* in all
                transcript isoforms mapping to the merged gene

            *masked*
                positions excluded from analyses as directed in an optional
                :term:`mask file`

    The following files are output, where `OUTBASE` is a name supplied
    by the user:

        OUTBASE_gene.positions
            Tab-delimited text file. Each line is a merged gene, and columns
            indicate the genomic coordinates and lengths of each of the position
            sets above.

        OUTBASE_transcript.positions
            Tab-delimited text file. Each line is a transcript, and columns
            indicate the genomic coordinates and lengths of each of the position
            sets above.

        OUTBASE_gene_REGION.bed
             `BED`_ files showing position sets for `REGION`,
             where `REGION` is one of *exon*, *CDS*, *UTR5*, and *UTR3* or
             *masked*. These contain the same information in
             ``OUTBASE_gene.positions``, but can be visualized easily in a
             :term:`genome browser`


Count
    The :func:`count <do_count>` mode counts the number and density of
    read alignments in each sub-region (*exon*, *CDS*, *UTR5*, and *UTR3*)
    of each gene.


Chart
    The :func:`chart <do_chart>` mode takes output from one or more samples
    run under :func:`count <do_count>` mode and generates several tables and
    charts that provide broad overviews of the data.

See command-line help for each subprogram for details on each mode

See also
--------
:mod:`~plastid.bin.counts_in_region` script
    Calculate  the number and density  of :term:`read alignments <alignment>`
    covering any set of regions of interest, making no corrections for gene
    boundaries.
"""
import os
import sys
import argparse
import itertools
import copy
import inspect
import gc
import warnings

import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt
import numpy
import pandas as pd
import scipy.stats

from plastid.util.scriptlib.argparsers import AnnotationParser,\
                                              AlignmentParser,\
                                              MaskParser,\
                                              PlottingParser, \
                                              BaseParser

from plastid.util.scriptlib.help_formatters import format_module_docstring

from plastid.genomics.roitools import positions_to_segments, SegmentChain
from plastid.genomics.genome_hash import GenomeHash
from plastid.util.io.openers import opener, get_short_name, argsopener, read_pl_table
from plastid.util.io.filters import NameDateWriter
from plastid.util.services.sets import merge_sets
from plastid.util.services.decorators import skipdoc
from plastid.util.services.exceptions import DataWarning, FileFormatWarning
import numpy.ma as ma
from plastid.plotting.plots import scatterhist_xy, ma_plot, clean_invalid
from plastid.plotting.colors import process_black
from plastid.readers.bigbed import BigBedReader

printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))

#===============================================================================
# 'generate' subprogram
#===============================================================================


@skipdoc
def write_output_files(table, title, args):
    """Write gene info table from :py:func:`do_generate` to several output files:

    OUTBASE_gene.positions
        Tab-delimited text file. Each line is a merged gene, and columns
        indicate the genomic coordinates and lengths of each of the position
        sets above.

    OUTBASE_transcript.positions
        Tab-delimited text file. Each line is a transcript, and columns
        indicate the genomic coordinates and lengths of each of the position
        sets above.

    OUTBASE_gene_REGION.bed
         `BED`_ files showing position sets for `REGION`,
         where `REGION` is one of *exon*, *utr5*, *cds*, *utr3*, or
         *masked*. These contain the same information in
         ``OUTBASE_gene.positions``, but can be visualized easily in a
         :term:`genome browser`


    Parameters
    ----------
    table : :class:`pandas.DataFrame`
        Gene info table made in :py:func:`do_generate`

    title : str
        Title ("gene" or "transcript")

    args : :py:class:`argparse.Namespace`
        Command-line arguments
    """
    keys = ("utr5", "utr3", "cds", "masked", "exon")
    bed_columns = ["%s_bed" % K for K in keys]

    bedfiles = {X: argsopener("%s_%s_%s.bed" % (args.outbase, title, X), args) for X in keys}

    for _, row in table[bed_columns].iterrows():
        for k in keys:
            bedfiles[k].write(row["%s_bed" % k])

    for k in bedfiles:
        bedfiles[k].close()

    pos_out = argsopener("%s_%s.positions" % (args.outbase, title), args)
    table.to_csv(
        pos_out,
        sep          = "\t",
        header       = True,
        index        = False,
        na_rep       = "nan",
        float_format = "%.8f",
        columns      = [
            "region", "exon", "utr5", "cds", "utr3", "masked", "exon_unmasked",
            "transcript_ids"
        ]) # yapf: disable
    pos_out.close()


def merge_genes(tx_ivcs):
    """Merge genes whose transcripts share exons into a combined, "merged" gene

    Parameters
    ----------
    tx_ivcs : dict
        Dictionary mapping unique transcript IDs to |Transcripts|


    Returns
    -------
    dict
        Dictionary mapping raw gene names to the names of the merged genes
    """
    dout = {}
    exondicts = {"+": {}, "-": {}}

    # backmap exons to genes as tuples
    printer.write("Mapping exons to genes ...")
    for txid in tx_ivcs.keys():
        my_segmentchain = tx_ivcs[txid]
        my_gene = my_segmentchain.get_gene()
        chrom = my_segmentchain.spanning_segment.chrom
        strand = my_segmentchain.spanning_segment.strand
        for my_iv in my_segmentchain:
            start = my_iv.start
            end = my_iv.end
            # separate exons by chromosome and strand to reduce
            # downstream comparisons in merging
            if chrom not in exondicts[strand]:
                exondicts[strand][chrom] = {}
            try:
                exondicts[strand][chrom][(start, end)].append(my_gene)
            except KeyError:
                exondicts[strand][chrom][(start, end)] = [my_gene]

    for strand in exondicts:
        for chrom in exondicts[strand]:
            exondicts[strand][chrom] = {K: set(V) for K, V in exondicts[strand][chrom].items()}
            printer.write("Flattening genes on %s(%s) ..." % (chrom, strand))
            gene_groups = merge_sets(exondicts[strand][chrom].values(), printer=printer)

            for group in gene_groups:
                merged_name = ",".join(sorted(group))
                for gene in group:
                    dout[gene] = merged_name

    printer.write("Flattened to %s groups." % len(dout))
    return dout


def process_partial_group(transcripts, mask_hash, printer):
    """Correct boundaries of merged genes, as described in :func:`do_generate`

    Parameters
    ----------
    transcripts : dict
        Dictionary mapping unique transcript IDs to |Transcripts|.
        This set should be complete in the sense that it should contain
        all transcripts that have any chance of mutually overlapping
        each other (e.g. all on same chromosome and strand).

    mask_hash : |GenomeHash|
        |GenomeHash| of regions to exclude from analysis


    Returns
    -------
    :class:`pandas.DataFrame`
        Table of merged gene positions

    :class:`pandas.DataFrame`
        Table of adjusted transcript positions

    :class:`dict`
        Dictionary mapping raw gene names to merged gene names
    """
    gene_table = {
        "region"          : [],
        "transcript_ids"  : [],
        "exon_unmasked"   : [],
        "exon"            : [],
        "masked"          : [],
        "utr5"            : [],
        "cds"             : [],
        "utr3"            : [],
        "exon_bed"        : [],
        "utr5_bed"        : [],
        "cds_bed"         : [],
        "utr3_bed"        : [],
        "masked_bed"      : [],
    } # yapf: disable

    # data table for transcripts
    transcript_table = {
        "region"          : [],
        "exon"            : [],
        "utr5"            : [],
        "cds"             : [],
        "utr3"            : [],
        "masked"          : [],
        "exon_unmasked"   : [],
        "transcript_ids"  : [],
        "exon_bed"        : [],
        "utr5_bed"        : [],
        "cds_bed"         : [],
        "utr3_bed"        : [],
        "masked_bed"      : [],
    } # yapf: disable

    keycombos = list(itertools.permutations(("utr5", "cds", "utr3"), 2))

    # merge genes that share exons & write output
    printer.write("Collapsing genes that share exons ...")
    merged_genes = merge_genes(transcripts)

    # remap transcripts to merged genes
    # and vice-versa
    merged_gene_tx = {}
    tx_merged_gene = {}
    printer.write("Mapping transcripts to merged genes...")
    for txid in transcripts:
        my_tx = transcripts[txid]
        my_gene = my_tx.get_gene()
        my_merged = merged_genes[my_gene]
        tx_merged_gene[txid] = my_merged
        try:
            merged_gene_tx[my_merged].append(txid)
        except KeyError:
            merged_gene_tx[my_merged] = [txid]

    # flatten merged genes
    printer.write("Flattening merged genes, masking positions, and labeling subfeatures ...")
    for n, (gene_id, my_txids) in enumerate(merged_gene_tx.items()):
        if n % 1000 == 0 and n > 0:
            printer.write("    %s genes ..." % n)

        my_gene_positions = []
        chroms = []
        strands = []
        for my_txid in my_txids:
            my_segmentchain = transcripts[my_txid]
            chroms.append(my_segmentchain.chrom)
            strands.append(my_segmentchain.strand)
            my_gene_positions.extend(my_segmentchain.get_position_list())

            try:
                assert len(set(chroms)) == 1
            except AssertionError:
                printer.write(
                    "Skipping gene %s which contains multiple chromosomes: %s" %
                    (gene_id, ",".join(chroms))
                )

            try:
                assert len(set(strands)) == 1
            except AssertionError:
                printer.write(
                    "Skipping gene %s which contains multiple strands: %s" %
                    (gene_id, ",".join(strands))
                )

        my_gene_positions = set(my_gene_positions)
        gene_ivc_raw = SegmentChain(
            *positions_to_segments(chroms[0], strands[0], my_gene_positions)
        )
        gene_table["region"].append(gene_id)
        gene_table["transcript_ids"].append(",".join(sorted(my_txids)))
        gene_table["exon_unmasked"].append(gene_ivc_raw)

    printer.write("    %s genes total." % (n + 1))

    # mask genes
    printer.write("Masking positions and labeling subfeature positions ...")
    gene_hash = GenomeHash(gene_table["exon_unmasked"], do_copy=False)

    for n, (gene_id, gene_ivc_raw) in enumerate(zip(gene_table["region"],
                                                    gene_table["exon_unmasked"])):
        if n % 2000 == 0:
            printer.write("    %s genes ..." % n)

        my_chrom = gene_ivc_raw.spanning_segment.chrom
        my_strand = gene_ivc_raw.spanning_segment.strand

        masked_positions = []
        nearby_genes = gene_hash[gene_ivc_raw]

        # don't mask out positions from identical gene
        gene_ivc_raw_positions = gene_ivc_raw.get_position_set()
        nearby_genes = [X for X in nearby_genes if X.get_position_set() != gene_ivc_raw_positions]
        for gene in nearby_genes:
            masked_positions.extend(gene.get_position_list())

        nearby_masks = mask_hash[gene_ivc_raw]
        for mask in nearby_masks:
            masked_positions.extend(mask.get_position_list())

        masked_positions = set(masked_positions)

        gene_positions_raw = gene_ivc_raw.get_position_set()
        mask_ivc_positions = gene_positions_raw & masked_positions
        total_mask_ivc = SegmentChain(
            *positions_to_segments(my_chrom, my_strand, mask_ivc_positions),
            ID=gene_id
        )
        gene_table["masked"].append(total_mask_ivc)
        gene_table["masked_bed"].append(total_mask_ivc.as_bed())

        gene_post_mask = gene_positions_raw - masked_positions
        gene_post_mask_ivc = SegmentChain(
            *positions_to_segments(my_chrom, my_strand, gene_post_mask),
            ID=gene_id
        )
        gene_table["exon"].append(gene_post_mask_ivc)
        gene_table["exon_bed"].append(gene_post_mask_ivc.as_bed())

        masked_positions = total_mask_ivc.get_position_set()
        tmp_positions = {
            "utr5" : set(),
            "cds"  : set(),
            "utr3" : set(),
        } # yapf: disable
        txids = sorted(merged_gene_tx[gene_id])
        chrom = gene_post_mask_ivc.chrom
        strand = gene_post_mask_ivc.strand

        # pool transcript positions
        for txid in txids:
            transcript = transcripts[txid]

            # yapf: disable
            utr5pos = transcript.get_utr5().get_position_set()
            cdspos  = transcript.get_cds().get_position_set()
            utr3pos = transcript.get_utr3().get_position_set()

            tmp_positions["utr5"] |= utr5pos
            tmp_positions["cds"]  |= cdspos
            tmp_positions["utr3"] |= utr3pos
            # yapf: enable

        # eliminate positions in which CDS & UTRs overlap from each transcript
        for txid in txids:
            transcript = transcripts[txid]
            transcript_positions = {
                "utr5": transcript.get_utr5().get_position_set(),
                "cds" : transcript.get_cds().get_position_set(),
                "utr3": transcript.get_utr3().get_position_set(),
            } # yapf: disable

            for key1, key2 in keycombos:
                transcript_positions[key1] -= tmp_positions[key2]
                transcript_positions[key1] -= masked_positions

            transcript_table["region"].append(txid)

            # all unmasked positions
            my_chain = SegmentChain(
                *positions_to_segments(
                    chrom,
                    strand,
                    transcript.get_position_set() - masked_positions
                ),
                ID=txid
            )
            transcript_table["exon"].append(str(my_chain))
            transcript_table["exon_bed"].append(my_chain.as_bed())

            # all uniquely-labeled unmasked positions
            for k, v in transcript_positions.items():
                my_chain = SegmentChain(*positions_to_segments(chrom, strand, v), ID=txid)
                transcript_table[k].append(str(my_chain))
                transcript_table["%s_bed" % k].append(my_chain.as_bed())

            total_mask_ivc.attr["ID"] = txid
            transcript_table["masked"].append(str(total_mask_ivc))
            transcript_table["masked_bed"].append(total_mask_ivc.as_bed())
            transcript_table["exon_unmasked"].append(str(transcript))
            transcript_table["transcript_ids"].append(txid)

        tmp_positions2 = copy.deepcopy(tmp_positions)
        for k1, k2 in keycombos:
            tmp_positions[k1] -= tmp_positions2[k2]
            tmp_positions[k1] -= masked_positions

        for k in (tmp_positions.keys()):
            my_chain = SegmentChain(
                *positions_to_segments(chrom, strand, tmp_positions[k]), ID=gene_id
            )
            gene_table[k].append(str(my_chain))
            gene_table["%s_bed" % k].append(my_chain.as_bed())

    printer.write("    %s genes total." % (n + 1))

    # cast SegmentChains/Transcripts to strings to keep numpy from unpacking them
    conversion_keys = ["exon", "utr5", "cds", "utr3", "masked", "exon_unmasked"]
    for k in conversion_keys:
        gene_table[k] = [str(X) for X in gene_table[k]]
        transcript_table[k] = [str(X) for X in transcript_table[k]]

    gene_df = pd.DataFrame(gene_table)
    gene_df.sort_values(["region"], inplace=True)

    transcript_df = pd.DataFrame(transcript_table)
    transcript_df.sort_values(["region"], inplace=True)

    return gene_df, transcript_df, merged_genes


def do_generate(args, annotation_parser, mask_parser):
    """Generate gene position files from gene annotations.

     1. Genes whose transcripts share exons are first collapsed into merged
        genes.

     2. Within merged genes, all positions are classified. All positions are
        included in a set called *exon*. All positions that appear as coding
        regions in all transcripts (i.e. are never part of a 5'UTR or 3'UTR)
        included in a set called *CDS*. Similarly, all positions that appear
        as 5' UTR or 3' UTR in all transcripts are included in sets called
        *UTR5* or *UTR3*, respectively.

     3. Genomic positions that are overlapped by multiple merged genes are
        excluded from the position sets for those genes.

     4. If a :term:`mask file` is supplied, positions annotated in the mask file
        are also excluded

     5. Output is given as a series of `BED`_ files and a `positions` file
        containing the same data.

    Parameters
    ----------
    args : :py:class:`argparse.Namespace`
        command-line arguments for ``generate`` subprogram
    """
    # variables for transcript <-> merged gene mapping
    transcripts = {}
    merged_genes = {}

    # data table for merged genes
    gene_table = pd.DataFrame({
            "region"          : [],
            "transcript_ids"  : [],
            "exon_unmasked"   : [],
            "exon"            : [],
            "masked"          : [],
            "utr5"            : [],
            "cds"             : [],
            "utr3"            : [],
            "exon_bed"        : [],
            "utr5_bed"        : [],
            "cds_bed"         : [],
            "utr3_bed"        : [],
            "masked_bed"      : [],
        }) # yapf: disable

    # data table for transcripts
    transcript_table = pd.DataFrame({
        "region"          : [],
        "exon"            : [],
        "utr5"            : [],
        "cds"             : [],
        "utr3"            : [],
        "exon_bed"        : [],
        "utr5_bed"        : [],
        "cds_bed"         : [],
        "utr3_bed"        : [],
        "masked"          : [],
        "exon_unmasked"   : [],
        "transcript_ids"  : [],
        "masked_bed"      : [],
    }) # yapf: disable

    # data
    is_sorted = (args.sorted == True) or \
                (args.tabix == True) or \
                (args.annotation_format == "BigBed")

    annotation_message = (
        "`cs` relies upon relationships between transcripts and genes "
        "to collapse transcripts to genes for quantitation. "
        "Gene-transcript relationships are not generally preserved in BED "
        "or BigBed files, and a `gene_id` column could not be found in the "
        "input data. This may yield nonsensical results in the output. "
        "Consider either (1) using a GTF2 or GFF3 file or (2) creating an "
        "extended BED or BigBed file with a `gene_id` column."
    )

    if args.annotation_format == "BED":
        if not isinstance(args.bed_extra_columns, list) \
            or 'gene_id' not in args.bed_extra_columns:
            warnings.warn(annotation_message, FileFormatWarning)

    elif args.annotation_format == "BigBed":
        reader = BigBedReader(args.annotation_files[0])
        if 'gene_id' not in reader.extension_fields:
            warnings.warn(annotation_message, FileFormatWarning)

    source = iter(
        annotation_parser.get_transcripts_from_args(args, printer=printer)
    )
    mask_hash = mask_parser.get_genome_hash_from_args(args)

    # loop conditions
    last_chrom = None
    do_loop = True

    # to save memory, we process one chromosome at a time if input file is sorted
    # knowing that at that moment all transcript parts are assembled
    while do_loop == True:
        try:
            tx = next(source)
        except StopIteration:
            do_loop = False

        try:
            # if chromosome is completely processed or EOF
            if (is_sorted and tx.spanning_segment.chrom != last_chrom) \
                or do_loop == False:
                if do_loop == True:
                    source = itertools.chain([tx], source)

                if last_chrom is not None or do_loop == False:
                    printer.write("Merging genes on chromosome/contig '%s'" % last_chrom)
                    my_gene_table, my_transcript_table, my_merged_genes = process_partial_group(
                        transcripts,
                        mask_hash,
                        printer,
                    )
                    gene_table = pd.concat((gene_table, my_gene_table), axis=0)
                    transcript_table = pd.concat((transcript_table, my_transcript_table), axis=0)
                    merged_genes.update(my_merged_genes)

                del transcripts
                gc.collect()
                del gc.garbage[:]
                transcripts = {}

                # reset last chrom
                last_chrom = tx.spanning_segment.chrom

            # otherwise, remember transcript
            else:
                transcripts[tx.get_name()] = tx

        # exit gracefully if no transcripts found
        except UnboundLocalError:
            pass

    # write output
    printer.write("Writing output ...")

    merged_fn = "%s_merged.txt" % args.outbase
    number_merged = len(set(merged_genes.values()))
    printer.write(
        "Collapsed %s genes to %s merged groups. Writing to %s" %
        (len(merged_genes), number_merged, merged_fn)
    )
    fout = argsopener(merged_fn, args, "w")
    for gene, merged_name in sorted(merged_genes.items()):
        fout.write("%s\t%s\n" % (gene, merged_name))

    fout.close()

    printer.write("Writing gene table and BED files ...")
    write_output_files(gene_table, "gene", args)
    printer.write("Writing transcript summary table and BED files ...")
    write_output_files(transcript_table, "transcript", args)
    printer.write("Done!")


#===============================================================================
# 'count' subprogram
#===============================================================================


def do_count(args, alignment_parser):
    """Count the number and density covering each merged gene in an annotation
    made made using the `generate` subcommand).

    Parameters
    ----------
    args : :py:class:`argparse.Namespace`
        command-line arguments for ``count`` subprogram
    """
    # we expect many zero-lenght segmentchains, so turn these off for now
    warnings.filterwarnings(
        "ignore",
        ".*zero-length SegmentChain.*",
    )

    keys = ("exon", "utr5", "cds", "utr3")
    column_order = ["region"]
    gene_positions = read_pl_table(args.position_file)

    # read count files
    ga = alignment_parser.get_genome_array_from_args(args, printer=printer)
    total_counts = ga.sum()

    normconst = 1000.0 * 1e6 / total_counts

    printer.write("Dataset has %s counts in it." % total_counts)
    printer.write("Tallying genes ...")

    dtmp = {"region": []}
    for x in keys:
        for y in ("reads", "length", "rpkm"):
            label = "%s_%s" % (x, y)
            dtmp[label] = []
            column_order.append(label)

    for i, name in enumerate(gene_positions["region"]):
        dtmp["region"].append(name)
        if i % 500 == 0:
            printer.write("Processed %s genes ..." % i)

        for k in keys:
            ivc = SegmentChain.from_str(gene_positions[k][i])
            total = sum(ivc.get_counts(ga))
            length = ivc.length
            rpkm = (normconst * total / length) if length > 0 else numpy.nan
            dtmp["%s_reads" % k].append(total)
            dtmp["%s_length" % k].append(length)
            dtmp["%s_rpkm" % k].append(rpkm)

    fout = argsopener("%s.txt" % args.outbase, args, "w")
    dtmp = pd.DataFrame(dtmp)
    dtmp.to_csv(
        fout,
        sep     = "\t",
        header  = True,
        index   = False,
        columns = column_order,
        na_rep  = "nan",
        float_format="%.8f"
    ) # yapf: disable

    fout.close()
    printer.write("Done.")


#===============================================================================
# 'chart' subprogram
#===============================================================================


@skipdoc
def read_count_file(fh, genes_to_include=None):
    """Read rows selected by gene name from a count file

    Parameters
    ----------
    fname : str
        Name of output file from cs.py

    genes_to_include : list or None
        List of genes to include in output.
        If `None`, all are included

    Returns
    -------
    :class:`pandas.DataFrame`

        Count data
    """
    df = read_pl_table(fh)
    if genes_to_include is not None:
        import_mask = numpy.array(
            [True if X in genes_to_include else False for X in df["region"]]
        )
        return df[import_mask]
    else:
        return df


@skipdoc
def get_bin_mask_by_summed_key(rep_a, rep_b, bins, key="exon_reads"):
    """Bins genes into groups based upon the summed counts in samples `rep_a`
    and `rep_b`

    Parameters
    ----------
    rep_a : :class:`pandas.DataFrame`
        replicate a, from :func:`read_count_file`

    rep_b : :class:`pandas.DataFrame`
        replicate b, from :func:`read_count_file`

    bins : :class:`numpy.ndarray`
        sequence of bins, specified as floors

    key : str, optional
        quantity on which to perform binning (Default: `exon_reads`)

    Returns
    -------
    dict[bin]
        = numpy boolean mask
    """
    my_sums = rep_a[key] + rep_b[key]
    bins = sorted(bins)
    dout = dict.fromkeys(bins)
    for k in dout.keys():
        dout[k] = numpy.tile(False, len(my_sums))
    for n, my_sum in enumerate(my_sums):
        for i in range(len(bins) - 1):
            bin1 = bins[i]
            bin2 = bins[i + 1]
            if my_sum >= bin1 and my_sum < bin2:
                dout[bin1][n] = True
        if my_sum >= bins[-1]:
            dout[bins[-1]][n] = True
    return dout


@skipdoc
def get_nonzero_either_mask(vector_a, vector_b):
    """Returns a numpy array of boolean values indicating where values in two
    vectors are both greater than zero.

    Parameters
    ----------
    vector_a : numpy.ndarray
        Array of counts or RPKM

    vector_b : numpy.ndarray
        Array of counts or RPKM

    Returns
    -------
    numpy.ndarray
        Boolean array that is `True` where both `vector_a` and `vector_b`
        have values greater than zero, and `False` elsewhere.
    """
    return (vector_a > 0) & (vector_b > 0)


@skipdoc
def get_short_samplename(inp):
    """Creates a sample legend label from a sample filename by removing
    everything including and after ``".txt"``

    Parameters
    ----------
    inp : str
        filename

    Returns
    -------
    str
    """
    return get_short_name(inp, separator=os.path.sep, terminator=".txt")


def do_scatter(
        x,
        y,
        count_mask,
        plot_parser,
        args,
        pearsonr = None,
        xlabel   = None,
        ylabel   = None,
        title    = None,
        min_x    = 10**-3,
        min_y    = 10**-3
): # yapf: disable
    """Scatter plot helper for cs `chart` subprogram

    Parameters
    ----------
    x, y : :class:`numpy.ndarray`
        Data to plot

    count_mask : :class:`numpy.ndarray`
        Threshold mask

    args : :class:`~argparse.Namespace`
        Command-line arguments

    pearsonr : float
        Pearson's r of the two samples

    xlabel : str or None, optional
        If not `None`, an x-axis label

    ylabel : str or None, optional
        If not `None`, a y-axis label

    min_x,min_y : float
        value to which low x- or y-values will respectively be truncated

    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Formatted figure
    """
    fig = plot_parser.get_figure_from_args(args)
    xm = x[count_mask]
    ym = y[count_mask]

    xb, yb = clean_invalid(x[~count_mask], y[~count_mask], min_x=min_x, min_y=min_y)

    _, axdict = scatterhist_xy(xm, ym, log="xy", label=">= 128 counts")
    _, _ = scatterhist_xy(
        xb,
        yb,
        axes  = axdict,
        color = process_black,
        label = "< 128 counts",
        mask_invalid=False,
    ) # yapf: disable

    text_kwargs = {
        "horizontalalignment": "left",
        "verticalalignment": "baseline",
        "transform": axdict["main"].transAxes
    }

    mainax = axdict["main"]
    mainax.legend(loc="lower right", frameon=False)

    if pearsonr is not None:
        mainax.text(0.02, 0.9, "Pearson r**2 >= 128: %0.4e" % pearsonr**2, **text_kwargs)

    if xlabel is not None:
        mainax.set_xlabel(xlabel)

    if ylabel is not None:
        mainax.set_ylabel(ylabel)

    if title is not None:
        plt.suptitle(title)

    mainax.set_xlim(min_x / 5.0, mainax.get_xlim()[1])
    mainax.set_ylim(min_y / 5.0, mainax.get_ylim()[1])

    return fig


def do_chart(args, plot_parser):
    """Produce a set of charts comparing multiple samples pairwise.

    Charts include histograms of log2 fold changes and scatter plots with
    correlation coefficients, both generated for raw count and RPKM data.

    Parameters
    ----------
    args : :py:class:`argparse.Namespace`
        command-line arguments for ``chart`` subprogram
    """
    plot_parser.set_style_from_args(args)

    outbase = args.outbase
    bins = numpy.array(args.bins)
    figformat = args.figformat

    # read input files
    printer.write("Reading input files: %s ..." % ", ".join(args.infiles))

    samples = { get_short_samplename(X) : read_count_file(opener(X),args.list_of_regions)\
                                          for X in args.infiles }

    # Define some variables for later
    sample_names = sorted(samples.keys())
    comparisons = [X for X in itertools.combinations(sample_names, 2) if X[0] != X[1]]
    colors = plot_parser.get_colors_from_args(args, len(comparisons))
    binkeys = tuple(["%s_reads" % k for k in args.regions])

    comparison_labels = sorted(["%s_vs_%s" % (X, Y) for X, Y in comparisons])
    bigtable = {}.fromkeys(comparison_labels)
    corrcoef_by_bin_table = {}.fromkeys(comparison_labels)

    # ki, kj = names of samples i and j
    # vi, vj = data of samples i and j
    for ki, kj in comparisons:
        try:
            assert (samples[ki]["region"] == samples[kj]["region"]).all()
        except AssertionError:
            printer.write(
                "Mismatched line entries for samples %s and %s. "
                "Were different gene lists used for counting?"
                % (ki, kj)
            )

        vi = samples[ki]
        vj = samples[kj]
        printer.write("Comparing %s to %s:" % (ki, kj))

        label = "%s_vs_%s" % (ki, kj)

        bigtable[label] = {}
        corrcoef_by_bin_table[label] = {}

        for binkey in binkeys:
            bigtable[label][binkey] = {K: copy.deepcopy({}) for K in args.metrics}
            corrcoef_by_bin_table[label][binkey] = {K: copy.deepcopy({}) for K in args.metrics}

        for region in args.regions:
            region_counts = "%s_reads" % region
            count_mask = (vi[region_counts].values + vj[region_counts].values >= 128)

            for metric in args.metrics:
                region_metric = "%s_%s" % (region, metric)
                printer.write("    -%s across %s for all >=128 ..." % (metric, region))

                # divide into regions >=128 and <128 and plot
                viover = vi[region_metric][count_mask].values
                vjover = vj[region_metric][count_mask].values

                viunder = vi[region_metric][~count_mask].values
                vjunder = vj[region_metric][~count_mask].values

                pearsonr = scipy.stats.pearsonr(viover, vjover)[0]
                spearmanr = scipy.stats.spearmanr(viover, vjover)[0]

                # log2 fold change stats
                log2_ratios = numpy.log2(numpy.ma.masked_invalid(vjover / viover))
                log2_mean = log2_ratios.mean()
                log2_std = log2_ratios.std()
                min_diff = 2**(3 * log2_std)
                num_genes_log2 = (~log2_ratios.mask).sum()

                # ma plot
                try:
                    ma_title = "%s vs %s (%s %s)" % (ki, kj, region, metric)
                    fig = plot_parser.get_figure_from_args(args)
                    _, axdict = ma_plot(
                        viunder,
                        vjunder,
                        color   = process_black,
                        label   = "< 128 counts",
                        axes    = plt.gca(),
                        kdalpha = 0.2,
                        title   = ma_title) # yapf: disable
                    _, _ = ma_plot(
                        viover,
                        vjover,
                        axes    = axdict,
                        label   = ">= 128 counts",
                        kdalpha = 0.8) # yapf: disable

                    mainax = axdict["main"]
                    mainax.set_xlabel("%s %s" % (ki, metric))
                    mainax.legend(loc="upper right", frameon=False)

                    text_kwargs = {
                        "horizontalalignment": "right",
                        "verticalalignment": "baseline",
                        "linespacing": 1.6,
                        "transform": mainax.transAxes,
                    }

                    plot_text = "\n".join(
                        [
                            "sample mean: %0.3e" % log2_mean,
                            "sample stdev: %0.3e" % log2_std,
                            "3-sigma fold change: %0.2f" % min_diff,
                            "regions_counted: %s" % count_mask.sum(),
                        ]
                    )

                    mainax.text(0.96, 0.04, plot_text, **text_kwargs)

                    plt.savefig(
                        "%s_ma_%s_%s_%s.%s" % (outbase, label, region, metric, figformat),
                        bbox_inches="tight"
                    )
                    plt.close()
                except Exception as e:
                    warnings.warn(
                        "Could not make MA plot for samples %s and %s. Error message:\n    %s" %
                        (ki, kj, e.message), DataWarning
                    )

                # scatter plot
                try:
                    scatter_title = "%s vs %s (%s %s)" % (ki, kj, region, metric)
                    do_scatter(
                        vi[region_metric].values,
                        vj[region_metric].values,
                        count_mask,
                        plot_parser,
                        args,
                        pearsonr = pearsonr,
                        xlabel   = ki,
                        ylabel   = kj,
                        title    = scatter_title
                    ) # yapf: disable

                    plt.savefig(
                        "%s_scatter_%s_%s_%s.%s" % (outbase, label, region, metric, figformat),
                        bbox_inches="tight"
                    )
                    plt.close()
                except ValueError as e:
                    warnings.warn(
                        "Could not make scatter plot for samples %s and %s. Error message:\n    %s"
                        % (ki, kj, e.message), DataWarning
                    )

                # TODO: make these tables into dataframes. this scheme is difficult to read

                # add entries to bigtable for export later
                # yapf: disable
                bigtable[label][region_counts][metric]["pearsonr"]  = pearsonr
                bigtable[label][region_counts][metric]["spearmanr"] = spearmanr
                bigtable[label][region_counts][metric]["log2_mean"] = log2_mean
                bigtable[label][region_counts][metric]["log2_std"]  = log2_std
                bigtable[label][region_counts][metric]["2**3sigma"] = 2**(3 * log2_std)
                bigtable[label][region_counts][metric][
                    "num_genes_128plus"] = count_mask.sum()
                bigtable[label][region_counts][metric][
                    "num_genes_log2"] = num_genes_log2
                # yapf: enable

                # do bin-by-bin counting
                printer.write("    -%s across %s by bin ..." % (metric, region))
                bin_masks = get_bin_mask_by_summed_key(vi, vj, bins=args.bins, key=region_counts)
                for my_bin, bin_mask in sorted(bin_masks.items()):

                    bin_vec_i = vi[region_metric][bin_mask]
                    bin_vec_j = vj[region_metric][bin_mask]

                    # make sure there are genes in bin before attempting calculations
                    if len(bin_vec_i) > 0:
                        nonzero_binmask = get_nonzero_either_mask(bin_vec_i, bin_vec_j)
                        bin_vi_log2 = bin_vec_i[nonzero_binmask]
                        bin_vj_log2 = bin_vec_j[nonzero_binmask]
                        my_logs = numpy.log2(bin_vj_log2 / bin_vi_log2)
                        my_logmean = numpy.mean(my_logs)
                        my_logstd = numpy.std(my_logs)
                        if len(bin_vec_i) > 2:
                            my_pearsonr = scipy.stats.pearsonr(bin_vec_i, bin_vec_j)[0]
                            my_spearmanr = scipy.stats.spearmanr(bin_vec_i, bin_vec_j)[0]
                        else:
                            my_spearmanr = numpy.nan
                            my_pearsonr = numpy.nan
                    else:
# yapf: disable
                        # fill with dummy values
                        my_logs      = numpy.array([])
                        my_logmean   = numpy.nan
                        my_logstd    = numpy.nan
                        my_spearmanr = numpy.nan
                        my_pearsonr  = numpy.nan

                    corrcoef_by_bin_table[label][region_counts][metric][my_bin] = {}
                    corrcoef_by_bin_table[label][region_counts][metric][my_bin]["pearsonr"]     = my_pearsonr
                    corrcoef_by_bin_table[label][region_counts][metric][my_bin]["spearmanr"]    = my_spearmanr
                    corrcoef_by_bin_table[label][region_counts][metric][my_bin]["genes_in_bin"] = sum(bin_mask)
                    corrcoef_by_bin_table[label][region_counts][metric][my_bin]["log2_genes_in_bin"] = sum(nonzero_binmask)
                    corrcoef_by_bin_table[label][region_counts][metric][my_bin]["log2_mean"]    = my_logmean
                    corrcoef_by_bin_table[label][region_counts][metric][my_bin]["log2_std"]     = my_logstd

    # export big (non-binned) table
    printer.write("Writing tables ...")

    bigtable_out = argsopener("%s_bigtable.txt" % args.outbase, args, "w")
    stats = (
        "num_genes_128plus",  # 0
        "pearsonr",           # 1
        "spearmanr",          # 2
        "num_genes_log2",     # 3
        "log2_mean",          # 4
        "log2_std",           # 5
        "2**3sigma",          # 6
    )
# yapf: enable

    header = ["#region", "metric", "statistic"]
    header += [X for X in comparison_labels]

    bigtable_out.write("\t".join(header) + "\n")
    for region in binkeys:
        for metric in args.metrics:
            for stat in stats:
                ltmp = [region, metric, stat]
                for key in comparison_labels:
                    ltmp.append(bigtable[key][region][metric][stat])
                ltmp = [str(X) for X in ltmp]
                bigtable_out.write("\t".join(ltmp) + "\n")

    bigtable_out.close()

    # export binned data table and make bin-by-bin plots
    bintable_out = argsopener("%s_bintable.txt" % args.outbase, args, "w")

    region_metrics = ["%s_%s" % (X, Y) for X in args.regions for Y in args.metrics]
    stats = [
        "genes_in_bin",       # 0
        "pearsonr",           # 1
        "spearmanr",          # 2
        "",                   # 3
        "log2_genes_in_bin",  # 4
        "log2_mean",          # 5
        "log2_std",           # 6
        "",                   # 7
    ] # yapf: disable
    for region_metric in region_metrics:
        plt.close()
        fig = plot_parser.get_figure_from_args(args)
        plt.semilogx(basex=2)
        plt.title("Correlation coefficients by bin for %s" % region_metric)
        plt.xlabel("Bin")
        plt.ylabel("Spearman rho")

        region, metric = region_metric.split("_")
        region_counts = "%s_reads" % region
        bintable_out.write("%s\n\n" % region_metric)

        try:
            for n, label in enumerate(comparison_labels):
                corrcoefs = []
                bintable_out.write("%s\t\t\t\t\t\t\t\t" % label)
                bintable_out.write("\n")
                bintable_out.write("bin\t" + "\t".join(stats) + "\n")
                for my_bin in bins:
                    ltmp = [my_bin]
                    for stat in stats:
                        ltmp.append(
                            corrcoef_by_bin_table[label][region_counts][metric][my_bin].get(
                                stat, ""
                            )
                        )
                    ltmp = [str(X) for X in ltmp]
                    bintable_out.write("\t".join(ltmp) + "\n\n")
                    corrcoefs.append(
                        corrcoef_by_bin_table[label][region_counts][metric][my_bin]["spearmanr"]
                    )
                corrcoefs = ma.masked_invalid(corrcoefs)
                plt.plot(
                    bins[~corrcoefs.mask], corrcoefs[~corrcoefs.mask], label=label, color=colors[n]
                )
            plt.legend(loc="lower right")
            plt.savefig(
                "%s_corrcoef_by_bin_%s_%s.%s" % (outbase, region_metric, label, figformat),
                bbox_inches="tight"
            )
        except Exception as e:
            warnings.warn(
                "Could not plot correlation-by-bin plot for '%s', '%s'" % (outbase, region_metric),
                DataWarning
            )

        bintable_out.write("\n\n")
    bintable_out.close()

    printer.write("Done.")


#===============================================================================
# main program body
#===============================================================================


def main(argv=sys.argv[1:]):
    """Command-line program

    Parameters
    ----------
    argv : list, optional
        A list of command-line arguments, which will be processed
        as if the script were called from the command line if
        :func:`main` is called directly.

        Default: `sys.argv[1:]`. The command-line arguments, if the script is
        invoked from the command line
    """
    al = AlignmentParser(disabled=["normalize"])
    an = AnnotationParser()
    mp = MaskParser()
    pl = PlottingParser()
    bp = BaseParser()

    # yapf: disable
    alignment_file_parser  = al.get_parser()
    annotation_file_parser = an.get_parser()
    mask_file_parser       = mp.get_parser()
    plotting_parser        = pl.get_parser()
    base_parser            = bp.get_parser()

    generator_help = "Create unambiguous position file from GFF3 annotation"
    generator_desc = format_module_docstring(do_generate.__doc__)

    counter_help = "Count reads in unambiguous gene positions"
    counter_desc = format_module_docstring(do_count.__doc__)

    chart_help = "Produce charts comparing reads between samples"
    chart_desc = format_module_docstring(do_chart.__doc__)

    parser = argparse.ArgumentParser(
        description     = format_module_docstring(__doc__),
        formatter_class = argparse.RawDescriptionHelpFormatter
    )
    subparsers = parser.add_subparsers(
        title       = "subcommands",
        description = "choose one of the following",
        dest        = "program"
    )
    gparser = subparsers.add_parser(
        "generate",
        help            = generator_help,
        description     = generator_desc,
        parents         = [base_parser, annotation_file_parser, mask_file_parser],
        formatter_class = argparse.RawDescriptionHelpFormatter,
    )
    cparser = subparsers.add_parser(
        "count",
        help            = counter_help,
        description     = counter_desc,
        parents         = [base_parser, alignment_file_parser],
        formatter_class = argparse.RawDescriptionHelpFormatter,
    )
    pparser = subparsers.add_parser(
        "chart",
        help            = chart_help,
        description     = chart_desc,
        parents         = [base_parser, plotting_parser],
        formatter_class = argparse.RawDescriptionHelpFormatter
    )

    gparser.add_argument(
        "outbase",
        metavar ="outbase",
        type    = str,
        help    = "Basename for output files"
    )

    cparser.add_argument(
        "position_file",
        type    = str,
        metavar = "file.positions",
        help    = (
            "File assigning positions to genes or transcripts "
            "(made using 'generate' subcommand)"
        )
    )
    cparser.add_argument("outbase", type=str, help="Basename for output files")

    pparser.add_argument(
        "-i",
        "--in",
        nargs="+",
        type=str,
        dest="infiles",
        help="input files, made by 'count' subprogram")
    pparser.add_argument(
        "--bins",
        nargs="+",
        type=int,
        default=(0, 32, 64, 128, 256, 512, 1024, 2048, 4096),
        help="Bins into which features are partitioned based on counts")
    pparser.add_argument(
        "--regions",
        nargs="+",
        type=str,
        default=("exon", "utr5", "cds", "utr3"),
        help="Regions to compare (default: exon, utr5, cds, utr3)")
    pparser.add_argument(
        "--metrics",
        nargs="+",
        type=str,
        default=("rpkm", "reads"),
        help="Metrics to compare (default: rpkm, reads)")
    pparser.add_argument(
        "list_of_regions",
        type=str,
        metavar='gene_list.txt',
        nargs="?",
        default=None,
        help=(
            "Optional. File listing regions (genes or transcripts), one per "
            "line, to include in comparisons. If not given, all genes are "
            "included."
        )
    )
    pparser.add_argument("outbase", type=str, help="Basename for output files")

    args = parser.parse_args(argv)
    bp.get_base_ops_from_args(args)

    if args.program == "generate":
        #generate position file
        do_generate(args, an, mp)

    elif args.program == "count":
        #use position file to count gene expression in infiles
        do_count(args, al)

    elif args.program == "chart":
        #use count files to generate a family of charts and tables
        do_chart(args, pl)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
from numpy.testing import suppress_warnings
from plastid.util.services.exceptions import DataWarning

#===============================================================================
# Warnings suppression
#
# Use within bodies of test functions as e.g. `with sup_data: foo`
# Do NOT use as function decorators, as these will hose `yield` tests
#===============================================================================

sup_data = suppress_warnings()
sup_data.filter(category=DataWarning)



"""Fast reader for `BigWig`_ files, implemented as a Python binding for the C
library of Jim Kent's utilties for the UCSC genome browser.


See also
--------
`Kent2010 <http://dx.doi.org/10.1093/bioinformatics/btq351>`_
    Description of BigBed and BigWig formats. Especially see supplemental data.

`Source repository for Kent utilities <https://github.com/ENCODE-DCC/kentUtils.git>`_
    The header files are particularly useful.
"""
from plastid.readers.bbifile cimport lm, bbiFile, _BBI_Reader, bits32, Bits, bbiSummaryType, get_lm
from plastid.genomics.roitools cimport GenomicSegment
from plastid.genomics.c_common cimport _GeneratorWrapper

#===============================================================================
# Externs from Kent utilties
#===============================================================================

cdef extern from "<bigWig.h>":
    cdef struct bigWigValsOnChrom:
        bigWigValsOnChrom *next
        char   *chrom
        long   chromSize
        long   bufSize     # size of allocated buffer
        double *valBuf     # value for each base on chrom. Zero where no data
        Bits   *covBuf     # a bit for each base with data

    cdef struct bbiInterval:
        bbiInterval * next
        bits32 start, end
        double val

    bbiFile     * bigWigFileOpen(char *fileName)

    bbiInterval * bigWigIntervalQuery(
        bbiFile *bwf,
        char *chrom,
        bits32 start,
        bits32 end,
        lm *lm
    )

    bigWigValsOnChrom *bigWigValsOnChromNew()

    void bigWigValsOnChromFree(bigWigValsOnChrom **pChromVals)

    bint bigWigValsOnChromFetchData(
        bigWigValsOnChrom *chromVals,
        char *chrom,
        bbiFile *bigWig
    )

    double bigWigSingleSummary(
        bbiFile *bwf,
        char *chrom,
        int start,
        int end,
        bbiSummaryType summaryType,
        double defaultVal
    )


#===============================================================================
# Python class
#===============================================================================


cdef class BigWigReader(_BBI_Reader):
    cdef double fill
    cdef double _sum

    cdef double _summarize(self, GenomicSegment roi, bbiSummaryType type_)
    cdef double c_sum(self)
    cdef bigWigValsOnChrom * c_get_chromosome_counts(self, str chrom)




#!/usr/bin/env python
"""Library components for writing command-line scripts

Package overview
================

    =================================================    =========================
    **Package module**                                   **Contents**
    -------------------------------------------------    -------------------------
    :py:mod:`~plastid.util.scriptlib.argparsers`          :class:`~argparse.ArgumentParser` objects for reading and processing various genomics file types 
    :py:mod:`~plastid.util.scriptlib.help_formatters`     Utilities to reformat module docstrings for use as command-line help text
    :py:mod:`~plastid.util.scriptlib.template`            Template for writing a command-line script using :data:`plastid`
    =================================================    =========================    
"""
